from typing import Optional, Union
from unittest import TestCase
from datetime import datetime, date
from jsonclasses import jsonclass
from jsonclasses.fdef import FType, FStore
from jsonclasses.excs import ValidationException
from tests.classes.linked_author import LinkedAuthor
from tests.classes.linked_article import LinkedArticle
from tests.classes.linked_user import LinkedUser
from tests.classes.linked_product import LinkedProduct
from tests.classes.linked_customer import LinkedCustomer
from tests.classes.linked_profile import LinkedProfile
from tests.classes.new_union import NewUnion, NewRUnion, NewOUnion


class TestAutoTypes(TestCase):

    def test_auto_generates_required_str(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestStr:
            val: str
        object = TestStr()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_str(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalStr:
            val: Optional[str]
        object = TestOptionalStr()
        object.validate()

    def test_auto_generates_required_str_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestStrStrType:
            val: 'str'
        object = TestStrStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_str_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalStrStrType:
            val: 'Optional[str]'
        object = TestOptionalStrStrType()
        object.validate()

    def test_auto_generates_required_int(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestInt:
            val: int
        object = TestInt()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_int(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalInt:
            val: Optional[int]
        object = TestOptionalInt()
        object.validate()

    def test_auto_generates_required_int_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestIntStrType:
            val: 'int'
        object = TestIntStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_int_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalIntStrType:
            val: 'Optional[int]'
        object = TestOptionalIntStrType()
        object.validate()

    def test_auto_generates_required_float(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestFloat:
            val: float
        object = TestFloat()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_float(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalfloat:
            val: Optional[float]
        object = TestOptionalfloat()
        object.validate()

    def test_auto_generates_required_float_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestFloatStrType:
            val: 'float'
        object = TestFloatStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_float_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalFloatStrType:
            val: 'Optional[float]'
        object = TestOptionalFloatStrType()
        object.validate()

    def test_auto_generates_required_bool(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestBool:
            val: bool
        object = TestBool()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_bool(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalbool:
            val: Optional[bool]
        object = TestOptionalbool()
        object.validate()

    def test_auto_generates_required_bool_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestboolStrType:
            val: 'bool'
        object = TestboolStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_bool_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalboolStrType:
            val: 'Optional[bool]'
        object = TestOptionalboolStrType()
        object.validate()

    def test_auto_generates_required_date(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class Testdate:
            val: date
        object = Testdate()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_date(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionaldate:
            val: Optional[date]
        object = TestOptionaldate()
        object.validate()

    def test_auto_generates_required_date_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestdateStrType:
            val: 'date'
        object = TestdateStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_date_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionaldateStrType:
            val: 'Optional[date]'
        object = TestOptionaldateStrType()
        object.validate()

    def test_auto_generates_required_datetime(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class Testdatetime:
            val: datetime
        object = Testdatetime()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_datetime(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionaldatetime:
            val: Optional[datetime]
        object = TestOptionaldatetime()
        object.validate()

    def test_auto_generates_required_datetime_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestdatetimeStrType:
            val: 'datetime'
        object = TestdatetimeStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_datetime_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionaldatetimeStrType:
            val: 'Optional[datetime]'
        object = TestOptionaldatetimeStrType()
        object.validate()

    def test_auto_generates_required_list(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class Testlist:
            val: list[str]
        object = Testlist()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_list(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class TestOptionallist:
            val: Optional[list[str]]
        object = TestOptionallist()
        object.validate()

    def test_auto_generates_required_list_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class TestlistStrType:
            val: 'list[str]'
        object = TestlistStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_list_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class TestOptionallistStrType:
            val: 'Optional[list[str]]'
        object = TestOptionallistStrType()
        object.validate()

    def test_auto_generates_required_list_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class Testlist:
            val: list[str]
        object = Testlist()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_list_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionallist:
            val: Optional[list[str]]
        object = TestOptionallist()
        object.validate()

    def test_auto_generates_required_list_with_str_type_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestlistStrType:
            val: 'list[str]'
        object = TestlistStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_list_with_str_type_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionallistStrType:
            val: 'Optional[list[str]]'
        object = TestOptionallistStrType()
        object.validate()

    def test_auto_generates_required_dict(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class Testdict:
            val: dict[str, str]
        object = Testdict()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_dict(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class TestOptionaldict:
            val: Optional[dict[str, str]]
        object = TestOptionaldict()
        object.validate()

    def test_auto_generates_required_dict_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class TestdictStrType:
            val: 'dict[str, str]'
        object = TestdictStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_dict_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen_2')
        class TestOptionaldictStrType:
            val: 'Optional[dict[str, str]]'
        object = TestOptionaldictStrType()
        object.validate()

    def test_auto_generates_required_dict_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class Testdict:
            val: dict[str, str]
        object = Testdict()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_dict_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionaldict:
            val: Optional[dict[str, str]]
        object = TestOptionaldict()
        object.validate()

    def test_auto_generates_required_dict_with_str_type_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestdictStrType:
            val: 'dict[str, str]'
        object = TestdictStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_dict_with_str_type_capitalized(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionaldictStrType:
            val: 'Optional[dict[str, str]]'
        object = TestOptionaldictStrType()
        object.validate()

    def test_auto_generates_required_instance(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestinstanceReferenced:
            val: Optional[str]

        @jsonclass(class_graph='test_marker_auto_gen')
        class Testinstance:
            val: TestinstanceReferenced
        object = Testinstance()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_instance(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestoptionalinstanceReferenced:
            val: Optional[str]

        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalInstance:
            val: Optional[TestoptionalinstanceReferenced]
        object = TestOptionalInstance()
        object.validate()

    def test_auto_generates_required_instance_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestinstanceReferencedWithStrType:
            val: Optional[str]

        @jsonclass(class_graph='test_marker_auto_gen')
        class TestinstanceWithStrType:
            val: 'TestinstanceReferencedWithStrType'
        object = TestinstanceWithStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_instance_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestoptionalinstanceReferencedWithStrType:
            val: Optional[str]

        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalInstanceWithStrType:
            val: 'Optional[TestoptionalinstanceReferencedWithStrType]'
        object = TestOptionalInstanceWithStrType()
        object.validate()

    def test_auto_generates_required_union(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestUnion:
            val: Union[str, bool]
        object = TestUnion()
        self.assertFalse(object.is_valid)

    def test_auto_generates_optional_union(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalUnion:
            val: Optional[Union[str, bool]]
        object = TestOptionalUnion()
        object.validate()

    def test_auto_generates_required_union_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestUnionStrType:
            val: 'Union[str, int]'
        object = TestUnionStrType()
        self.assertRaises(ValidationException, object.validate)

    def test_auto_generates_optional_union_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalUnionStrType:
            val: 'Optional[Union[int, float]]'
        object = TestOptionalUnionStrType()
        object.validate()

    def test_auto_generates_required_union_with_dict_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestUnionDictType:
            val: 'Union[dict[str, int], int]'
        cfield = TestUnionDictType.cdef.field_named('val')
        utypes = cfield.fdef.raw_union_types
        self.assertEqual(utypes[0].fdef.ftype, FType.DICT)
        self.assertEqual(utypes[1].fdef.ftype, FType.INT)

    def test_auto_generates_optional_union_with_dict_with_str_type(self):
        @jsonclass(class_graph='test_marker_auto_gen')
        class TestOptionalUnionDictType:
            val: 'Optional[Union[dict[str, bool], float]]'
        object = TestOptionalUnionDictType()
        object.validate()

    def test_auto_generates_1_to_1_links(self):
        profile_field = LinkedUser.cdef.field_named('profile')
        self.assertEqual(profile_field.fdef.ftype,
                         FType.INSTANCE)
        self.assertEqual(profile_field.fdef.fstore,
                         FStore.FOREIGN_KEY)
        self.assertEqual(profile_field.fdef.foreign_key, 'user')
        self.assertEqual(profile_field.fdef.use_join_table, False)

        user_field = LinkedProfile.cdef.field_named('user')
        self.assertEqual(user_field.fdef.ftype,
                         FType.INSTANCE)
        self.assertEqual(user_field.fdef.fstore,
                         FStore.LOCAL_KEY)
        self.assertEqual(user_field.fdef.foreign_key, None)
        self.assertEqual(user_field.fdef.use_join_table, None)

    def test_auto_generates_1_to_many_links(self):
        articles_field = LinkedAuthor.cdef.field_named('articles')
        self.assertEqual(articles_field.fdef.ftype,
                         FType.LIST)
        self.assertEqual(articles_field.fdef.fstore,
                         FStore.FOREIGN_KEY)
        self.assertEqual(articles_field.fdef.foreign_key, 'author')
        self.assertEqual(articles_field.fdef.use_join_table, False)

        author_field = LinkedArticle.cdef.field_named('author')
        self.assertEqual(author_field.fdef.ftype,
                         FType.INSTANCE)
        self.assertEqual(author_field.fdef.fstore,
                         FStore.LOCAL_KEY)
        self.assertEqual(author_field.fdef.foreign_key, None)
        self.assertEqual(author_field.fdef.use_join_table, None)

    def test_auto_generates_many_to_many_links(self):
        customer1 = LinkedCustomer(name='C1')
        customer2 = LinkedCustomer(name='C2')
        product1 = LinkedProduct(name='P1')
        product2 = LinkedProduct(name='P2')
        customer1.products.append(product1)
        customer1.products.append(product2)
        product1.customers.append(customer2)
        self.assertEqual(customer1.products, [product1, product2])
        self.assertEqual(product1.customers, [customer1, customer2])
        self.assertEqual(customer2.products, [product1])
        self.assertEqual(product2.customers, [customer1])

    def test_auto_generates_optional_with_new_union_none_syntax_str(self):
        obj = NewUnion()
        obj.validate()
        obj.name = '123'
        obj.validate()
        obj.name = 123
        self.assertRaises(ValidationException, obj.validate)

    def test_auto_generates_optional_union_with_new_union_syntax_str(self):
        obj = NewUnion()
        obj.validate()
        obj.code = '123'
        obj.validate()
        obj.code = 123
        obj.validate()
        obj.code = 123.5
        self.assertRaises(ValidationException, obj.validate)

    def test_auto_generates_required_union_with_new_union_syntax_str(self):
        obj = NewRUnion(items=['12'])
        self.assertRaises(ValidationException, obj.validate)
        obj.code = '123'
        obj.validate()
        obj.code = 123
        obj.validate()
        obj.code = 123.5
        self.assertRaises(ValidationException, obj.validate)

    def test_auto_generates_required_complicated_union_with_new_union_syntax_str(self):
        obj = NewRUnion(items=['12'], code='12')
        obj.validate()
        obj.items = {'a': 1, 'b': '2'}
        obj.validate()
        obj.items = [1]
        self.assertRaises(ValidationException, obj.validate)
        obj.items = {'a': 5.0}
        self.assertRaises(ValidationException, obj.validate)
        obj.items = None
        self.assertRaises(ValidationException, obj.validate)

    def test_auto_generates_optional_complicated_union_with_new_union_syntax_str(self):
        obj = NewOUnion(items=['12'])
        obj.validate()
        obj.items = {'a': 1, 'b': '2'}
        obj.validate()
        obj.items = [1]
        self.assertRaises(ValidationException, obj.validate)
        obj.items = {'a': 5.0}
        self.assertRaises(ValidationException, obj.validate)
        obj.items = None
        obj.validate()



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class LinkedOwner:
    id: str = types.str.primary.required
    permissions: list[LinkedPermission] = types.listof('LinkedPermission').linkedby('owner')


@jsonclass(
    can_read=types.getop.isobj(types.this.fval('owner'))
)
class LinkedPermission:
    id: str = types.str.primary.required
    owner: LinkedOwner = types.objof('LinkedOwner').linkto.required



"""module for tonextsec modifier."""
from __future__ import annotations
from datetime import datetime, timedelta
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToNextSecModifier(Modifier):
    """Change the second to the next second"""

    def transform(self, ctx: Ctx) -> Any:
        return (ctx.val+timedelta(seconds=1)).replace(microsecond=0) if type(ctx.val) is datetime else ctx.val



from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase
from tests.classes.format import Format


class TestFmt(TestCase):

    def test_fmt_takes_callable_format(self):
        f = Format(color1='123456')
        self.assertEqual(f.tojson()['color1'], '#123456')
        self.assertEqual(f.color1, '123456')

    def test_fmt_takes_types_format(self):
        f = Format(color2='123456')
        self.assertEqual(f.tojson()['color2'], '#123456')
        self.assertEqual(f.color2, '123456')



"""module for gt modifier."""
from __future__ import annotations
from typing import Callable, Union, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class GtModifier(Modifier):
    """Gt modifier validates value against min value."""

    def __init__(self, gt_value: int | float | Callable | Types):
        self.gt_value = gt_value

    def validate(self, ctx: Ctx) -> None:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        if is_number and ctx.val <= self.resolve_param(self.gt_value, ctx):
            ctx.raise_vexc('value is not greater than '
                           f'{self.gt_value}')



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleCode:
    code: Optional[str] = types.str.length(4)
    min_code: Optional[str] = types.str.minlength(4)
    max_code: Optional[str] = types.str.maxlength(8)

    l_code: Optional[list[int]] = types.listof(int).length(4)
    l_min_code: Optional[list[int]] = types.listof(int).minlength(4)
    l_max_code: Optional[list[int]] = types.listof(int).maxlength(8)

    c_code: Optional[str] = types.str.length(lambda: 4)
    c_min_code: Optional[str] = types.str.minlength(lambda: 4)
    c_max_code: Optional[str] = types.str.maxlength(lambda: 8)

    cl_code: Optional[list[int]] = types.listof(int).length(lambda: 4)
    cl_min_code: Optional[list[int]] = types.listof(int).minlength(lambda: 4)
    cl_max_code: Optional[list[int]] = types.listof(int).maxlength(lambda: 8)

    t_code: Optional[str] = types.str.length(types.default(4))
    t_min_code: Optional[str] = types.str.minlength(types.default(4))
    t_max_code: Optional[str] = types.str.maxlength(types.default(8))

    tl_code: Optional[list[int]] = types.listof(int).length(types.default(4))
    tl_min_code: Optional[list[int]] = types.listof(int).minlength(types.default(4))
    tl_max_code: Optional[list[int]] = types.listof(int).maxlength(types.default(8))

    cd_code: Optional[str] = types.str.length(lambda: 4, lambda: 5)
    td_code: Optional[str] = types.str.length(types.default(4), types.default(4).add(1))



"""module for at modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class AtModifier(Modifier):
    """At modifier returns result with subscription index."""

    def __init__(self, index: Any | Callable | Types) -> None:
        self.index = index

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        return ctx.val[self.resolve_param(self.index, ctx)]



"""module for nonnull modifier."""
from .modifier import Modifier
from ..fdef import FDef, Nullability


class NonnullModifier(Modifier):
    """A nonnull modifier transforms None into empty library."""

    def define(self, fdef: FDef) -> None:
        fdef._collection_nullability = Nullability.NONNULL



"""module for modifier modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from ..fdef import FDef, FType
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class TypeModifier(Modifier):
    """Abstract modifier for checking object's type."""

    def __init__(self) -> None:
        self.cls: type = object
        self.ftype: FType = FType.ANY
        self.exact_type: bool = False

    def define(self, fdef: FDef) -> None:
        fdef._ftype = self.ftype

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        if self.exact_type:
            if type(ctx.val) is self.cls:
                return
        else:
            if isinstance(ctx.val, self.cls):
                return
        ctx.raise_vexc(f'value is not {self.cls.__name__}')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.hex_color import HexColor


class TestHexColor(TestCase):

    def test_hexcolor_removes_initial_pound(self):
        color = HexColor(hex_color='#01FD6F')
        self.assertEqual(color.hex_color, '01FD6F')

    def test_hexcolor_doesnt_raise(self):
        color = HexColor(hex_color='01FD6F')
        color.validate()

    def test_hexcolor_raises_if_value_is_not_valid(self):
        color = HexColor(hex_color='ZZZZZZ')
        with self.assertRaises(ValidationException) as context:
            color.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['hexColor'],
                         "value is not hex color string")



from __future__ import annotations
from typing import TYPE_CHECKING
from jsonclasses import jsonclass, types
if TYPE_CHECKING:
    from tests.classes.linked_student import LinkedStudent


@jsonclass
class LinkedSchool:
    name: str
    students: list[LinkedStudent] = types.nonnull.listof('LinkedStudent')



from jsonclasses import jsonclass, types


@jsonclass
class UserSOS:
    name: str = types.str.required
    age: int = types.int.setonsave(lambda x: x + 300).required


@jsonclass
class UserSOSZ:
    name: str = types.str.required
    age: int = types.int.setonsave(lambda: 500).required


@jsonclass
class UserFSOS:
    name: str = types.str.required
    age: int = types.int.fsetonsave(lambda x: (x or 0) + 100).required


@jsonclass
class UserTFSOS:
    name: str = types.str.required
    code: str = types.str.fsetonsave(types.randomdigits(4)).required



"""module for outputvalue modifier."""
from ..fdef import FDef, EnumOutput
from .modifier import Modifier


class OutputValueModifier(Modifier):
    """Output value modifier tweaks enum modifier's behavior."""

    def define(self, fdef: FDef) -> None:
        fdef._enum_output = EnumOutput.VALUE



from __future__ import annotations
from typing import Any
from jsonclasses import jsonclass, types

val = 0


def check_value() -> int:
    return val


def callback() -> None:
    global val
    val = val + 1


def callbackd(value: int) -> None:
    global val
    val = val + value


def callbackt(value: int, context: Any) -> None:
    global val
    val = val + value + context.val


@jsonclass
class OnwriteSingle:
    a: str = types.str.required
    b: int = types.int.onwrite(callback).required


@jsonclass
class OnwriteDouble:
    a: str = types.str.required
    b: int = types.int.onwrite(callbackd).required


@jsonclass
class OnwriteTriple:
    a: str = types.str.required
    b: int = types.int.onwrite(callbackt).required



from __future__ import annotations
from tests.classes.simple_calculation import SimpleCalculation
from unittest import TestCase

class TestSub(TestCase):


    def test_sub_with_int_value_sub_original_value(self):
        sub_int = SimpleCalculation(i_sub=8)
        self.assertEqual(sub_int.i_sub, 3)

    def test_sub_with_float_value_sub_original_value(self):
        sub_float = SimpleCalculation(f_sub=8.5)
        self.assertEqual(sub_float.f_sub, 6.0)

    def test_sub_subs_callable_value_to_original_value(self):
        sub_float = SimpleCalculation(c_sub=8.4)
        self.assertEqual(sub_float.c_sub, 5.9)

    def test_sub_subs_types_value_to_original_value(self):
        sub_float = SimpleCalculation(t_sub=8.4)
        self.assertEqual(sub_float.t_sub, 5.9)



"""module for writeonce modifier."""
from ..fdef import FDef, WriteRule
from .modifier import Modifier


class WriteonceModifier(Modifier):
    """Writeonce modifier marks a field as writeonce."""

    def define(self, fdef: FDef) -> None:
        fdef._write_rule = WriteRule.WRITE_ONCE



from __future__ import annotations
from jsonclasses.excs import ValidationException
from tests.classes.super_prefix import IsPrefixOf
from unittest import TestCase


class TestIsPrefixOf(TestCase):

    def test_is_prefix_of_validates_str_val_is_prefix_of_str_param(self):
        s_ipo = IsPrefixOf(s_ipo='unha')
        s_ipo.validate()

    def test_is_prefix_of_validates_list_val_of_int_is_prefix_of_list_param(self):
        loi_ipo = IsPrefixOf(loi_ipo=[1, 4, 5])
        loi_ipo.validate()

    def test_is_prefix_of_validates_list_val_of_str_is_prefix_of_list_param(self):
        los_ipo = IsPrefixOf(los_ipo=['a', 'd'])
        los_ipo.validate()

    def test_is_prefix_of_raises_str_val_is_not_prefix_of_str_param(self):
        s_ipo = IsPrefixOf(s_ipo='appy')
        with self.assertRaises(ValidationException) as context:
            s_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['sIpo'],
                         "prefix is not found")

    def test_is_prefix_of_raises_list_val_of_int_is_not_prefix_of_list_param(self):
        loi_ipo = IsPrefixOf(loi_ipo=[3, 2, 8])
        with self.assertRaises(ValidationException) as context:
            loi_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['loiIpo'],
                         "prefix is not found")

    def test_is_prefix_of_raises_list_val_of_str_is_not_prefix_of_list_param(self):
        los_ipo = IsPrefixOf(los_ipo=['f', 'g'])
        with self.assertRaises(ValidationException) as context:
            los_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['losIpo'],
                         "prefix is not found")

    def test_is_prefix_of_validates_str_val_is_prefix_of_callable_param(self):
        c_ipo = IsPrefixOf(c_ipo='unha')
        c_ipo.validate()

    def test_is_prefix_of_validates_list_val_of_int_is_prefix_of_callable_param(self):
        c_loi_ipo = IsPrefixOf(c_loi_ipo=[1, 4, 5])
        c_loi_ipo.validate()

    def test_is_prefix_of_validates_list_val_of_str_is_prefix_of_callable_param(self):
        c_los_ipo = IsPrefixOf(c_los_ipo=['a', 'd'])
        c_los_ipo.validate()

    def test_is_prefix_of_raises_str_val_is_not_prefix_of_callable_param(self):
        c_ipo = IsPrefixOf(c_ipo='appy')
        with self.assertRaises(ValidationException) as context:
            c_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['cIpo'],
                         "prefix is not found")

    def test_is_prefix_of_raises_list_val_of_int_is_not_prefix_of_callable_param(self):
        c_loi_ipo = IsPrefixOf(c_loi_ipo=[3, 2, 8])
        with self.assertRaises(ValidationException) as context:
            c_loi_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['cLoiIpo'],
                         "prefix is not found")

    def test_is_prefix_of_raises_list_val_of_str_is_not_prefix_of_callable_param(self):
        c_los_ipo = IsPrefixOf(c_los_ipo=['f', 'g'])
        with self.assertRaises(ValidationException) as context:
            c_los_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['cLosIpo'],
                         "prefix is not found")

    def test_is_prefix_of_validates_str_val_is_prefix_of_types_param(self):
        t_ipo = IsPrefixOf(t_ipo='unha')
        t_ipo.validate()

    def test_is_prefix_of_validates_list_val_of_int_is_prefix_of_types_param(self):
        t_loi_ipo = IsPrefixOf(t_loi_ipo=[1, 4, 5])
        t_loi_ipo.validate()

    def test_is_prefix_of_validates_list_val_of_str_is_prefix_of_types_param(self):
        t_los_ipo = IsPrefixOf(t_los_ipo=['a', 'd'])
        t_los_ipo.validate()

    def test_is_prefix_of_raises_str_val_is_not_prefix_of_types_param(self):
        t_ipo = IsPrefixOf(t_ipo='appy')
        with self.assertRaises(ValidationException) as context:
            t_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['tIpo'],
                         "prefix is not found")

    def test_is_prefix_of_raises_list_val_of_int_is_not_prefix_of_types_param(self):
        t_loi_ipo = IsPrefixOf(t_loi_ipo=[3, 2, 8])
        with self.assertRaises(ValidationException) as context:
            t_loi_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['tLoiIpo'],
                         "prefix is not found")

    def test_is_prefix_of_raises_list_val_of_str_is_not_prefix_of_types_param(self):
        t_los_ipo = IsPrefixOf(t_los_ipo=['f', 'g'])
        with self.assertRaises(ValidationException) as context:
            t_los_ipo.validate()
        self.assertEqual(context.exception.keypath_messages['tLosIpo'],
                         "prefix is not found")



"""module for dictof modifier."""
from __future__ import annotations
from typing import Any, Iterable, Collection
from ..fdef import FType
from ..jconf import JConf
from .collection_type_modifier import CollectionTypeModifier


class DictOfModifier(CollectionTypeModifier):
    """This modifier validates dict."""

    def __init__(self, raw_item_types: Any) -> None:
        super().__init__(raw_item_types)
        self.cls = dict
        self.ftype = FType.DICT

    def enumerator(self, value: dict[str, Any]) -> Iterable:
        return value.items()

    def empty_collection(self) -> Collection:
        return {}

    def append_value(self, i: str, v: Any, col: dict):
        col[i] = v

    def to_object_key(self, key: str, conf: JConf) -> str:
        return key

    def to_json_key(self, key: str, conf: JConf) -> str:
        return key



from __future__ import annotations
from tests.classes.super_round import SuperRound
from unittest import TestCase



class TestRound(TestCase):

    def test_round_transforms_value_with_float_point_greater_than_point_5_into_int_by_rounding(self):
        sr = SuperRound(round_value=19.8)
        self.assertEqual(sr.round_value, 20)

    def test_round_transforms_value_with_float_point_less_than_point_5_into_int_by_rounding(self):
        sr = SuperRound(round_value=19.4)
        self.assertEqual(sr.round_value, 19)

    def test_round_keeps_int_value_unchanged(self):
        sr = SuperRound(round_value=40)
        self.assertEqual(sr.round_value, 40)



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperMessage:
    name: str = types.str.vmsg(types.length(6), "MUST 666").required



from __future__ import annotations
from typing import TYPE_CHECKING
from jsonclasses import jsonclass, types
if TYPE_CHECKING:
    from tests.classes.linked_school import LinkedSchool


@jsonclass
class LinkedStudent:
    name: str
    school: LinkedSchool = types.linkto.objof('LinkedSchool').present



"""module for filter modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class FilterModifier(Modifier):
    """Filter modifier filters value."""

    def __init__(self, callback: Callable | Types) -> None:
        self.callback = callback

    def transform(self, ctx: Ctx) -> Any:
        from ..types import Types
        if not isinstance(ctx.val, list):
            return ctx.val
        if isinstance(self.callback, Types):
            result = []
            for v in ctx.val:
                tsfmd = self.callback.modifier.transform(ctx.nval(v))
                try:
                    self.callback.modifier.validate(ctx.nval(tsfmd))
                    result.append(v)
                except:
                    continue
            return result
        else:
            result = []
            for v in ctx.val:
                callbackv = self.callback(v)
                if callbackv is None or callbackv is True:
                    result.append(v)
            return result



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import AbstractJSONClassException
from tests.classes.abstract_object import (AbstractObject, MyObject,
                                           NonAbstractObject,
                                           DefaultNonAbstractObject)


class TestAbstractClass(TestCase):

    def test_abstract_class_cannot_be_initialized(self):
        with self.assertRaises(AbstractJSONClassException) as context:
            AbstractObject()
        self.assertEqual(
            context.exception.message,
            'AbstractObject is an abstract class and should not be '
            'initialized')

    def test_non_abstract_class_can_be_initialized(self):
        NonAbstractObject()

    def test_abstract_classs_non_abstract_subclass_can_be_initialized(self):
        MyObject()

    def test_abstract_class_class_is_not_abstract_by_default(self):
        DefaultNonAbstractObject()



from __future__ import annotations
from enum import Enum
from jsonclasses import jsonclass, jsonenum, types


@jsonenum
class Gender(Enum):
    MALE = 1
    FEMALE = 2


@jsonclass
class EnumUser:
    name: str = types.str.required
    gender: Gender = types.enum(Gender).required



"""module for embedded modifier."""
from ..fdef import FDef, FStore
from .modifier import Modifier


class EmbeddedModifier(Modifier):
    """This modifier marks value as embedded on the hosting object."""

    def define(self, fdef: FDef) -> None:
        fdef._fstore = FStore.EMBEDDED



"""module for enum modifier."""
from __future__ import annotations
from typing import Any, Union, TYPE_CHECKING
from enum import Enum
from ..fdef import (FType, FDef, EnumInput,
                                EnumOutput)
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class EnumModifier(Modifier):
    """Enum modifier validates value against provided enum type."""

    def __init__(self, enum_or_name: Union[type[Enum], str]):
        super().__init__()
        self.enum_or_name = enum_or_name

    def define(self, fdef: FDef) -> None:
        fdef._ftype = FType.ENUM
        fdef._raw_enum_class = self.enum_or_name
        if fdef._enum_input is None:
            fdef._enum_input = EnumInput.NAME
        if fdef._enum_output is None:
            fdef._enum_output = EnumOutput.NAME
        return

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if isinstance(self.enum_or_name, str):
            jconf = ctx.cdefowner.jconf
            enum_class = jconf.cgraph.fetch_enum(self.enum_or_name)
        else:
            enum_class = self.enum_or_name
        if isinstance(ctx.val, enum_class):
            return ctx.val
        enum_input = ctx.fdef.enum_input
        if enum_input.__contains__(EnumInput.VALUE):
            try:
                return enum_class(ctx.val)
            except ValueError:
                pass
        elif isinstance(ctx.val, str):
            if enum_input.__contains__(EnumInput.NAME):
                try:
                    return enum_class[ctx.val]
                except KeyError:
                    pass
            if enum_input.__contains__(EnumInput.LOWERCASE_NAME):
                try:
                    return enum_class[ctx.val.upper()]
                except KeyError:
                    pass
            ctx.raise_vexc('unknown enum value')
        else:
            ctx.raise_vexc('unknown enum value')

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return None
        if isinstance(self.enum_or_name, str):
            jconf = ctx.cdefowner.jconf
            enum_class = jconf.cgraph.fetch_enum(self.enum_or_name)
        else:
            enum_class = self.enum_or_name
        if not isinstance(ctx.val, enum_class):
            ctx.raise_vexc('invalid enum value')

    def tojson(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if ctx.fdef.enum_output == EnumOutput.VALUE:
            return ctx.val.value
        elif ctx.fdef.enum_output == EnumOutput.NAME:
            return ctx.val.name
        elif ctx.fdef.enum_output == EnumOutput.LOWERCASE_NAME:
            return ctx.val.name.lower()



from __future__ import annotations
from tests.classes.super_iterable import SuperIterable
from unittest import TestCase


class TestReplacer(TestCase):

    def test_replacer_keeps_str_value_if_no_replacement_is_found(self):
        s = SuperIterable(itssub="bbbccc")
        self.assertEqual(s.itssub, "bbbccc")

    def test_replacer_replaces_first_appearance(self):
        s = SuperIterable(itssub="dd qq 1 rr")
        self.assertEqual(s.itssub, "dd qq ABC rr")

    def test_replacer_replaces_multiple_appearances(self):
        s = SuperIterable(itssub="gg ss 2 dd 5 ee 8")
        self.assertEqual(s.itssub, "gg ss ABC dd ABC ee ABC")

    def test_replacer_keeps_callable_value_if_no_replacement_is_found(self):
        s = SuperIterable(c_itssub="bbbccc")
        self.assertEqual(s.c_itssub, "bbbccc")

    def test_replacer_keeps_types_value_if_no_replacement_is_found(self):
        s = SuperIterable(t_itssub="bbbccc")
        self.assertEqual(s.t_itssub, "bbbccc")



from __future__ import annotations
from unittest import TestCase
from tests.classes.cb_article import CBArticle
from tests.classes.cbm_article import CBMArticle
from tests.classes.cbo_article import CBOArticle


class TestDecoratedOnCreate(TestCase):

    def test_callback_are_called_for_new_objects_on_save(self):
        article = CBArticle(name='A', content='B')
        article.save()
        self.assertEqual(len(article.revisions), 1)
        self.assertEqual(article.revisions[0].name, 'First')

    def test_callback_are_not_called_for_existing_objects_on_save(self):
        article = CBArticle(name='A', content='B')
        setattr(article, '_is_new', False)
        article.save()
        self.assertEqual(len(article.revisions), 0)

    def test_multiple_callbacks_are_called_for_new_objects_on_save(self):
        article = CBMArticle(name='A', content='B')
        article.save()
        self.assertEqual(len(article.revisions), 1)
        self.assertEqual(article.revisions[0].name, 'First')
        self.assertEqual(article.content, 'UPDATED')

    def test_multiple_callbacks_are_not_called_for_existing_objects_on_save(self):
        article = CBMArticle(name='A', content='B')
        setattr(article, '_is_new', False)
        article.save()
        self.assertEqual(len(article.revisions), 0)
        self.assertEqual(article.content, 'B')

    def test_operator_can_be_passed_into_callback(self):
        article = CBOArticle(name='A', content='B')
        article.opby('operator 1')
        article.save()
        self.assertEqual(len(article.revisions), 1)
        self.assertEqual(article.revisions[0].name, 'operator 1')
        self.assertEqual(article.content, 'operator 1')



from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase
from tests.classes.super_datetime import SuperDateTime

class TestTonextyear(TestCase):

    def test_tonextyear_transforms_datetime_into_the_time_of_next_year(self):
        d = SuperDateTime(dtny=datetime(2021, 10, 11, 17, 37, 27, 446259))
        self.assertEqual(d.dtny, datetime(2022,1, 1, 0, 0))

    def test_tonextyear_transforms_date_into_the_time_of_next_year(self):
        d = SuperDateTime(dny=date(2021, 10, 11))
        self.assertEqual(d.dny, date(2022,1, 1))

    def test_tonextyear_does_not_transform_if_is_not_datetime_or_date(self):
        s = SuperDateTime(sny='12345')
        self.assertEqual(s.sny, '12345')



"""module for minlength modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class MinlengthModifier(Modifier):
    """Minlength modifier validates value against min length."""

    def __init__(self, minlength: int | Callable | Types) -> None:
        self.minlength = minlength

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, list) or type(ctx.val) is str:
            if len(ctx.val) < self.resolve_param(self.minlength, ctx):
                ctx.raise_vexc('length of value is not greater than or equal '
                            f'{self.minlength}')
        return ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleContact:
    email: Optional[str] = types.str.presentwithout(['phone_no', 'address'])
    phone_no: Optional[str] = types.str.presentwithout(['email', 'address'])
    address: Optional[str] = types.str.presentwithout(['phone_no', 'email'])



"""module for even modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class EvenModifier(Modifier):
    """Even modifier raises if int value is not even."""

    def validate(self, ctx: Ctx) -> None:
        if type(ctx.val) is int and ctx.val % 2 != 0:
            ctx.raise_vexc('value is not even')



"""module for authorization identity modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from inspect import signature
from ..fdef import FDef
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class AuthByModifier(Modifier):
    """Fields marked with auth by are used for authorization.
    """

    def __init__(self, checker: Types) -> None:
        self.checker = checker

    def define(self, fdef: FDef) -> None:
        fdef._auth_by = True
        fdef._auth_by_checker = self.checker



from __future__ import annotations
from datetime import datetime
from unittest import TestCase
from tests.classes.super_datetime import SuperDateTime

class TestTobomin(TestCase):

    def test_tobomin_transforms_datetime_into_the_time_of_beginning_of_minute(self):
        d = SuperDateTime(dtbm=datetime(2021, 10, 11, 17, 37, 27))
        self.assertEqual(d.dtbm, datetime(2021,10, 11, 17, 37))

    def test_tobomin_does_not_transform_if_is_not_datetime(self):
        s = SuperDateTime(stbm="12345")
        self.assertEqual(s.stbm, "12345")



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class CellphoneTitle:
    cellphone_name: str = types.str.required
    cellphone_title: str = types.str.totitle.required




from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class TransformName:
    name: Optional[str] = types.str.transform(lambda s: s + 'q')


@jsonclass
class CTransformName:
    name: Optional[str] = types.str.transform(lambda s, c: s + c.val)


@jsonclass
class TTransformName:
    age: Optional[int] = types.int.transform(types.add(5).mul(5))



"""module for random int modifier."""
from __future__ import annotations
from random import randrange
from typing import Any, TYPE_CHECKING, Callable, Union
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class RandomIntModifier(Modifier):
    """Random int modifier generates random int as result."""

    def __init__(self, min_value: int | float | Callable | Types,
                max_value: int | float | Callable | Types) -> None:
        self.min_value = min_value
        self.max_value = max_value

    def transform(self, ctx: Ctx) -> Any:
        min_value = self.resolve_param(self.min_value, ctx)
        max_value = self.resolve_param(self.max_value, ctx)
        if min_value < max_value:
            return randrange(min_value, max_value)
        else:
            message = f'{min_value} is not less than or equal {max_value}'
        ctx.raise_vexc(message)



"""This module defineds the JSON Class object mapping graph."""
from __future__ import annotations
from typing import Iterator, NamedTuple, Union, TYPE_CHECKING
from .isjsonclass import isjsonobject
from .excs import (UnlinkableJSONClassException,
                         JSONClassGraphMergeConflictException)
if TYPE_CHECKING:
    from .jobject import JObject


class OGraph:
    """The object graph is a graph containing JSON Class objects. It has two
    main usages. First, it's used for tracking and referencing objects within
    the same objects from a query result. Second, it's used for marking objects
    as handled when performing validating and serializing.
    """

    def __init__(self: OGraph,
                 objects: Union[list[JObject],
                                JObject, None] = None) -> None:
        self._maps: dict[str, dict[str, JObject]] = {}
        try:
            if isinstance(objects, list):
                for object in objects:
                    self.put(object)
            if isjsonobject(objects):
                self.put(objects)
        except UnlinkableJSONClassException:
            return

    def __iter__(self) -> Iterator[JObject]:
        lst: list[JObject] = []
        for table in self._maps.values():
            for obj in table.values():
                lst.append(obj)
        return lst.__iter__()

    def _object_map(self: OGraph,
                    name: str) -> dict[str, JObject]:
        if self._maps.get(name) is None:
            self._maps[name] = {}
        return self._maps[name]

    def _check_get_str_id(self: OGraph, object: JObject) -> str:
        primary_value = object._id
        if primary_value is None:
            if object.__class__.cdef.primary_field:
                has_field = True
            else:
                has_field = False
            raise UnlinkableJSONClassException(type(object), has_field)
        return str(primary_value)

    def put(self: OGraph, object: JObject) -> None:
        """Put an object into object graph. When conflict, this object
        overrides the existing one on the graph.
        """
        primary_value = self._check_get_str_id(object)
        class_name = object.__class__.__name__
        object_map = self._object_map(class_name)
        object_map[primary_value] = object

    def has(self: OGraph, object: JObject) -> bool:
        """Check whether an object is existing in the object graph.
        """
        return self.get(object) is not None

    def get(self: OGraph, object: JObject) -> JObject:
        """Get an object from the graph which matches the provided object.
        """
        primary_value = self._check_get_str_id(object)
        class_name = object.__class__.__name__
        object_map = self._object_map(class_name)
        return object_map.get(primary_value)

    def copy(self: OGraph) -> OGraph:
        """Get a copy of this object graph.
        """
        new_graph = OGraph()
        for key, map in self._maps.items():
            new_graph._maps[key] = {}
            for name, obj in map.items():
                new_graph._maps[key][name] = obj
        return new_graph


    def merged_graph(self: OGraph, graph2: OGraph) -> OGraph:
        """Get a new graph which is a combination of two graphs.
        """
        if self is graph2:
            return self
        graph = self.copy()
        for object in graph2:
            if not graph.has(object):
                graph.put(object)
            elif graph.get(object) is object:
                graph.put(object)
            else:
                raise JSONClassGraphMergeConflictException(
                    'multiple objects represent same object: ', object)
        for object in graph:
            object._graph = graph
        return graph



"""module for replace modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class ReplaceModifier(Modifier):
    """Replace modifier Replace  value."""

    def __init__(self, old: str | Callable | Types, new: str | Callable | Types) -> None:
        self.old = old
        self.new = new

    def transform(self, ctx: Ctx) -> Any:
        is_str = type(ctx.val) is str
        old = self.resolve_param(self.old, ctx)
        new = self.resolve_param(self.new, ctx)
        return ctx.val.replace(old, new) if is_str else ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleChart:
    name: Optional[str]
    partitions: dict[str, float] = types.dictof(types.float.max(1)).required



from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkedthru, types
if TYPE_CHECKING:
    from tests.classes.linked_default_post import LinkedDefaultPost


@jsonclass
class LinkedDefaultUser:
    name: str = types.str.default('Default User')
    posts: Annotated[list[LinkedDefaultPost], linkedthru('users')]



"""module for length modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class LengthModifier(Modifier):
    """Length modifier validate value against the provided length."""

    def __init__(self, minlength: int | Callable | Types, maxlength: int | Callable | Types | None = None) -> None:
        self.minlength = minlength
        self.maxlength = maxlength if maxlength is not None else minlength

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, list) or type(ctx.val) is str:
            minlength = self.resolve_param(self.minlength, ctx)
            if self.minlength == self.maxlength:
                maxlength = minlength
            else:
                maxlength = self.resolve_param(self.maxlength, ctx)
            if len(ctx.val) > maxlength or len(ctx.val) < minlength:
                if minlength != maxlength:
                    msg = f'length of value is not between {minlength} and {maxlength}'
                else:
                    msg = f'length of value is not {minlength}'
                ctx.raise_vexc(msg)
        return ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperInsertAt:
    s: Optional[str] = types.str.insertat(432, 3)
    l: Optional[list[str]] = types.listof(str).insertat('E', 1)

    cs: Optional[str] = types.str.insertat(lambda: 432, lambda: 3)
    ts: Optional[str] = types.str.insertat(types.default(432), types.default(3))



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleOrder:
    name: Optional[str]
    quantity: Optional[int] = types.int.default(1)



from __future__ import annotations
from enum import Enum
from jsonclasses import jsonclass, jsonenum, types


@jsonenum
class ValueGender(Enum):
    MALE = 1
    FEMALE = 2


@jsonclass
class ValueGenderUser:
    name: str = types.str.required
    gender: ValueGender = types.enum(ValueGender).inputvalue.outputvalue \
                               .required



from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase
from tests.classes.super_datetime import SuperDateTime

class TestTonextday(TestCase):

    def test_tonextday_transforms_datetime_into_the_time_of_next_day(self):
        d = SuperDateTime(dtnd=datetime(2021, 10, 11, 17, 37, 27, 446259))
        self.assertEqual(d.dtnd, datetime(2021,10, 12, 0, 0))

    def test_tonextday_transforms_date_into_the_time_of_next_day(self):
        d = SuperDateTime(dnd=date(2021, 10, 11))
        self.assertEqual(d.dnd, date(2021,10, 12))

    def test_tonextday_does_not_transform_if_is_not_datetime_or_date(self):
        s = SuperDateTime(snd='12345')
        self.assertEqual(s.snd, '12345')



from __future__ import annotations
from datetime import date
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_article import SimpleArticle
from tests.classes.simple_order import SimpleOrder
from tests.classes.simple_address import SimpleAddress
from tests.classes.simple_student import SimpleStudent
from tests.classes.simple_deadline import SimpleDeadline
from tests.classes.author import Author
from tests.classes.article import Article
from tests.classes.default_dict import DefaultDict
from tests.classes.linked_song import LinkedSong, LinkedSinger


class TestInitialize(TestCase):

    def test_initialize_simple_object_without_arguments(self):
        article = SimpleArticle()
        self.assertEqual(article._data_dict, {'title': None, 'content': None})

    def test_initialize_simple_object_with_arguments(self):
        article = SimpleArticle(title='Oi', content='Tik')
        self.assertEqual(article._data_dict, {'title': 'Oi', 'content': 'Tik'})

    def test_initialize_simple_object_with_types_default_values(self):
        order = SimpleOrder(name='Oi Tik')
        self.assertEqual(order._data_dict, {'name': 'Oi Tik', 'quantity': 1})

    def test_initialize_simple_object_with_assigned_default_values(self):
        student = SimpleStudent()
        self.assertEqual(student._data_dict, {'age': 20, 'graduated': False})

    def test_initialize_with_value_passed_in_rather_than_default_value(self):
        student = SimpleStudent(graduated=True, age=24)
        self.assertEqual(student._data_dict, {'age': 24, 'graduated': True})

    def test_initialize_do_not_accept_undefined_keys_by_default(self):
        with self.assertRaises(ValidationException) as context:
            SimpleArticle(dzimsikai='Ku Piang Hoê')
        self.assertTrue(len(context.exception.keypath_messages) == 1)
        self.assertEqual(context.exception.keypath_messages['dzimsikai'],
                         "key is not allowed")

    def test_initialize_underscore_key_cases_by_default(self):
        address = SimpleAddress(**{'lineOne': 'NgouOu', 'lineTwo': 'Sihai'})
        self.assertEqual(address._data_dict,
                         {'line_one': 'NgouOu', 'line_two': 'Sihai'})

    def test_initialize_triggers_transform(self):
        deadline = SimpleDeadline(ended_at='2021-06-30')
        self.assertEqual(deadline.ended_at, date(2021, 6, 30))

    def test_initialize_accepts_object_list(self):
        article = Article(title='Khi Sit', content='Ua Sim Lai Tsa Tiu E Tsai')
        author = Author(name='Hun', articles=[article])
        self.assertEqual(author.articles, [article])

    def test_initialize_accepts_object(self):
        author = Author(name='Kieng')
        article = Article(title='E Sai',
                          content='Tsê Tioh Si Kim Sieng Ua Ê Tsuê Ai',
                          author=author)
        self.assertEqual(author, article.author)

    def test_initialize_accepts_nested_keypaths_for_instances(self):
        author = Author(name='Kieng')
        article = Article(**{'title': 'E Sai',
                             'content': 'Tsê Tioh Si Kim Sieng Ua Ê Tsuê Ai',
                             'author': author,
                             'author.name': 'abc.def'})
        self.assertEqual(article.author.name, 'abc.def')

    def test_initialize_accepts_nested_keypaths_for_instances_in_lists(self):
        article = Article(title='T', content='C')
        author = Author(**{'name': 'Kieng', 'articles': [article], 'articles.0.title': 'QQQQQ'})
        self.assertEqual(author.articles[0].title, 'QQQQQ')

    def test_initialize_accepts_nested_keypaths_for_value_in_dicts(self):
        dct = DefaultDict(**{'value.b': '3', 'value.c': '4'})
        self.assertEqual(dct.value, {'a': '1', 'b': '3', 'c': '4'})

    def test_initialized_object_list_local_keys_can_be_accessed(self):
        song = LinkedSong()
        self.assertEqual(song.singers, [])
        self.assertEqual(song.singer_ids, [])

    def test_initialized_object_list_add_local_object_syncs_local_key(self):
        song = LinkedSong(id=1)
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song.singers = [singer1]
        self.assertEqual(song.singer_ids, [1])
        song.singers.append(singer2)
        self.assertEqual(song.singer_ids, [1, 2])

    def test_initialized_object_list_del_local_object_syncs_local_key(self):
        song = LinkedSong(id=1)
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song.singers = [singer1]
        self.assertEqual(song.singer_ids, [1])
        song.singers.append(singer2)
        self.assertEqual(song.singer_ids, [1, 2])
        song.singers.remove(singer2)
        self.assertEqual(song.singer_ids, [1])

    def test_initialized_object_list_assign_local_object_syncs_local_key(self):
        song = LinkedSong(id=1)
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        singer3 = LinkedSinger(id=3)
        singer4 = LinkedSinger(id=4)
        song.singers = [singer1]
        self.assertEqual(song.singer_ids, [1])
        song.singers.append(singer2)
        self.assertEqual(song.singer_ids, [1, 2])
        song.singers = [singer3, singer4]
        self.assertEqual(song.singer_ids, [3, 4])
        self.assertEqual(song.singers, [singer3, singer4])

    def test_initialized_object_list_add_local_key_syncs_local_object(self):
        song = LinkedSong(id=1)
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song.singers = [singer1, singer2]
        song.singer_ids.append(3)
        self.assertEqual(song.singers, [singer1, singer2])

    def test_initialized_object_list_del_local_key_syncs_local_object(self):
        song = LinkedSong(id=1)
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song.singers = [singer1, singer2]
        self.assertEqual(singer2.songs, [song])
        song.singer_ids.remove(2)
        self.assertEqual(song.singers, [singer1])
        self.assertEqual(singer2.songs, [])

    def test_initialized_object_list_assign_local_key_syncs_local_object(self):
        song = LinkedSong(id=1)
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        singer3 = LinkedSinger(id=3)
        singer4 = LinkedSinger(id=4)
        song.singers = [singer1]
        self.assertEqual(song.singer_ids, [1])
        song.singer_ids = [7, 8, 1, 9, 10]
        self.assertEqual(song.singers, [singer1])
        self.assertEqual(song.singer_ids, [7, 8, 1, 9, 10])



from __future__ import annotations
from typing import Any, NamedTuple
from unittest import TestCase
from jsonclasses.olist import OwnedList


class AddRecord(NamedTuple):
    olist: OwnedList
    idx: int
    val: Any


class DelRecord(NamedTuple):
    olist: OwnedList
    val: Any


class SorRecord(NamedTuple):
    olist: OwnedList


class Owner:

    def __init__(self):
        self.add_records: list[AddRecord] = []
        self.del_records: list[DelRecord] = []
        self.sor_records: list[SorRecord] = []

    def __olist_will_change__(self, olist: OwnedList) -> None:
        pass

    def __olist_add__(self, olist: OwnedList, idx: int, val: Any) -> None:
        self.add_records.append(AddRecord(olist, idx, val))

    def __olist_del__(self, olist: OwnedList, val: Any) -> None:
        self.del_records.append(DelRecord(olist, val))

    def __olist_sor__(self, olist: OwnedList) -> None:
        self.sor_records.append(SorRecord(olist))


class TestOwnedList(TestCase):

    def test_owned_list_get_notified_thru_append(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list.append(4)
        self.assertEqual(owned_list, [1, 2, 3, 4])
        self.assertEqual(owner.add_records, [AddRecord(owned_list, 3, 4)])

    def test_owned_list_get_notified_thru_extend(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list.extend([4, 5])
        self.assertEqual(owned_list, [1, 2, 3, 4, 5])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 3, 4),
            AddRecord(owned_list, 4, 5)])

    def test_owned_list_get_notified_thru_insert(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list.insert(0, 0)
        self.assertEqual(owned_list, [0, 1, 2, 3])
        self.assertEqual(owner.add_records, [AddRecord(owned_list, 0, 0)])

    def test_owned_list_get_notified_thru_insert_idx_overflow(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list.insert(100, 0)
        self.assertEqual(owned_list, [1, 2, 3, 0])
        self.assertEqual(owner.add_records, [AddRecord(owned_list, 3, 0)])

    def test_owned_list_get_notified_thru_remove(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list.remove(1)
        self.assertEqual(owned_list, [2, 3])
        self.assertEqual(owner.del_records, [DelRecord(owned_list, 1)])

    def test_owned_list_get_notified_thru_sort(self):
        owner = Owner()
        owned_list = OwnedList([1, 5, 3])
        owned_list.owner = owner
        owned_list.sort()
        self.assertEqual(owned_list, [1, 3, 5])
        self.assertEqual(owner.sor_records, [SorRecord(owned_list)])

    def test_owned_list_get_notified_thru_clear(self):
        owner = Owner()
        owned_list = OwnedList([1, 5, 3])
        owned_list.owner = owner
        owned_list.clear()
        self.assertEqual(owned_list, [])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 5),
            DelRecord(owned_list, 3)])

    def test_owned_list_get_notified_thru_pop(self):
        owner = Owner()
        owned_list = OwnedList([1, 5, 3])
        owned_list.owner = owner
        owned_list.pop()
        self.assertEqual(owned_list, [1, 5])
        self.assertEqual(owner.del_records, [DelRecord(owned_list, 3)])

    def test_owned_list_get_notified_thru_pop_argument(self):
        owner = Owner()
        owned_list = OwnedList([1, 5, 3])
        owned_list.owner = owner
        owned_list.pop(0)
        self.assertEqual(owned_list, [5, 3])
        self.assertEqual(owner.del_records, [DelRecord(owned_list, 1)])

    def test_owned_list_get_notified_thru_reverse(self):
        owner = Owner()
        owned_list = OwnedList([1, 5, 3])
        owned_list.owner = owner
        owned_list.reverse()
        self.assertEqual(owned_list, [3, 5, 1])
        self.assertEqual(owner.sor_records, [SorRecord(owned_list)])

    def test_owned_list_get_notified_thru_plus_equal_sign(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list += [4, 5]
        self.assertEqual(owned_list, [1, 2, 3, 4, 5])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 3, 4),
            AddRecord(owned_list, 4, 5)])

    def test_owned_list_get_notified_thru_multiply_equal_sign_gt_1(self):
        owner = Owner()
        owned_list = OwnedList([1, 2])
        owned_list.owner = owner
        owned_list *= 2
        self.assertEqual(owned_list, [1, 2, 1, 2])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 2, 1),
            AddRecord(owned_list, 3, 2)])
        self.assertEqual(owner.del_records, [])

    def test_owned_list_do_not_get_notified_thru_multiply_equal_sign_1(self):
        owner = Owner()
        owned_list = OwnedList([1, 2])
        owned_list.owner = owner
        owned_list *= 1
        self.assertEqual(owned_list, [1, 2])
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [])

    def test_owned_list_get_notified_thru_multiply_equal_lt_1(self):
        owner = Owner()
        owned_list = OwnedList([1, 2])
        owned_list.owner = owner
        owned_list *= 0
        self.assertEqual(owned_list, [])
        self.assertEqual(owner.add_records, [])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 2)])

    def test_owned_list_get_notified_thru_subscript_set(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list[0] = -1
        self.assertEqual(owned_list, [-1, 2, 3])
        self.assertEqual(owner.del_records, [DelRecord(owned_list, 1)])
        self.assertEqual(owner.add_records, [AddRecord(owned_list, 0, -1)])

    def test_owned_list_get_notified_thru_subscript_set_negative(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        owned_list[-1] = 9
        self.assertEqual(owned_list, [1, 2, 9])
        self.assertEqual(owner.del_records, [DelRecord(owned_list, 3)])
        self.assertEqual(owner.add_records, [AddRecord(owned_list, 2, 9)])

    def test_owned_list_raises_if_subscript_set_out_of_index_range(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        with self.assertRaisesRegex(IndexError,
                                    'list assignment index out of range'):
            owned_list[10] = -1

    def test_owned_list_raises_if_subscript_set_neg_out_of_index_range(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        with self.assertRaisesRegex(IndexError,
                                    'list assignment index out of range'):
            owned_list[-10] = -1

    def test_owned_list_get_notified_thru_subscript_slice_set(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[1:3] = [7, 8]
        self.assertEqual(owned_list, [1, 7, 8, 4, 5])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 2),
            DelRecord(owned_list, 3)])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 1, 7),
            AddRecord(owned_list, 2, 8)])

    def test_owned_list_get_notified_thru_subscript_slice_set_overflow(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[100:1] = [6, 7]
        self.assertEqual(owned_list, [1, 2, 3, 4, 5, 6, 7])
        self.assertEqual(owner.del_records, [])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 5, 6),
            AddRecord(owned_list, 6, 7)])

    def test_owned_list_get_notified_thru_subscript_slice_set_neg(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[-3:] = [7, 8, 9]
        self.assertEqual(owned_list, [1, 2, 7, 8, 9])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 4),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 2, 7),
            AddRecord(owned_list, 3, 8),
            AddRecord(owned_list, 4, 9)])

    def test_owned_list_get_notified_thru_subscript_slice_set_neg_ovfl(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[-999:1] = [-1, 0]
        self.assertEqual(owned_list, [-1, 0, 2, 3, 4, 5])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1)])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 0, -1),
            AddRecord(owned_list, 1, 0)])

    def test_owned_list_get_notified_thru_subscript_slice_set_step(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[0::2] = [7, 8, 9]
        self.assertEqual(owned_list, [7, 2, 8, 4, 9])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 0, 7),
            AddRecord(owned_list, 2, 8),
            AddRecord(owned_list, 4, 9)])

    def test_owned_list_get_notified_thru_subscript_slice_set_step_neg(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[-5::2] = [7, 8, 9]
        self.assertEqual(owned_list, [7, 2, 8, 4, 9])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 0, 7),
            AddRecord(owned_list, 2, 8),
            AddRecord(owned_list, 4, 9)])

    def test_owned_list_get_notified_thru_subscript_slice_set_step_err(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        with self.assertRaisesRegex(ValueError,
                                    ('attempt to assign sequence of size 3 to'
                                     ' extended slice of size 1')):
            owned_list[2::10] = [7, 8, 9]

    def test_owned_list_get_notified_thru_subscript_slice_set_step_ovfl(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[10::2] = []
        self.assertEqual(owned_list, [1, 2, 3, 4, 5])
        self.assertEqual(owner.del_records, [])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_sub_slice_set_step_neg_ovfl(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        owned_list[-100::2] = [7, 8, 9]
        self.assertEqual(owned_list, [7, 2, 8, 4, 9])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [
            AddRecord(owned_list, 0, 7),
            AddRecord(owned_list, 2, 8),
            AddRecord(owned_list, 4, 9)])

    def test_owned_list_get_notified_thru_subscript_del(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        del owned_list[0]
        self.assertEqual(owned_list, [2, 3])
        self.assertEqual(owner.del_records, [DelRecord(owned_list, 1)])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_subscript_del_negative(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        del owned_list[-1]
        self.assertEqual(owned_list, [1, 2])
        self.assertEqual(owner.del_records, [DelRecord(owned_list, 3)])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_raises_if_subscript_del_out_of_index_range(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        with self.assertRaisesRegex(IndexError,
                                    'list assignment index out of range'):
            del owned_list[10]

    def test_owned_list_raises_if_subscript_del_neg_out_of_index_range(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3])
        owned_list.owner = owner
        with self.assertRaisesRegex(IndexError,
                                    'list assignment index out of range'):
            del owned_list[-10]

    def test_owned_list_get_notified_thru_subscript_slice_del(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[1:3]
        self.assertEqual(owned_list, [1, 4, 5])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 2),
            DelRecord(owned_list, 3)])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_subscript_slice_del_overflow(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[100:1]
        self.assertEqual(owned_list, [1, 2, 3, 4, 5])
        self.assertEqual(owner.del_records, [])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_subscript_slice_del_neg(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[-3:]
        self.assertEqual(owned_list, [1, 2])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 4),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_subscript_slice_del_neg_ovfl(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[-999:1]
        self.assertEqual(owned_list, [2, 3, 4, 5])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1)])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_subscript_slice_del_step(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[0::2]
        self.assertEqual(owned_list, [2, 4])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_subscript_slice_del_step_neg(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[-5::2]
        self.assertEqual(owned_list, [2, 4])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_subscript_slice_del_step_ovfl(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[10::2]
        self.assertEqual(owned_list, [1, 2, 3, 4, 5])
        self.assertEqual(owner.del_records, [])
        self.assertEqual(owner.add_records, [])

    def test_owned_list_get_notified_thru_sub_slice_del_step_neg_ovfl(self):
        owner = Owner()
        owned_list = OwnedList([1, 2, 3, 4, 5])
        owned_list.owner = owner
        del owned_list[-100::2]
        self.assertEqual(owned_list, [2, 4])
        self.assertEqual(owner.del_records, [
            DelRecord(owned_list, 1),
            DelRecord(owned_list, 3),
            DelRecord(owned_list, 5)])
        self.assertEqual(owner.add_records, [])



"""This is an internal module."""
from __future__ import annotations
from typing import cast, Any, Callable, Optional, TYPE_CHECKING
from enum import Enum, Flag
from .rtypes import rnamedtypes, rtypes
from .isjsonclass import isjsonclass
from .jobject import JObject
if TYPE_CHECKING:
    from .types import Types
    from .cdef import CDef


class FType(Enum):
    """Defined field types of jsonclass fields.
    """

    STR = 'str'
    INT = 'int'
    FLOAT = 'float'
    BOOL = 'bool'
    DATE = 'date'
    DATETIME = 'datetime'
    ENUM = 'enum'
    LIST = 'list'
    DICT = 'dict'
    INSTANCE = 'instance'
    ANY = 'any'
    UNION = 'union'


class FSubtype(Enum):
    """Defined field subtypes of jsonclass fields.
    """
    FILE = 'file'
    MONGOID = 'mongoid'


class FStore(Enum):
    """Defined field storage types of jsonclass fields.
    """

    EMBEDDED = 'embedded'
    LOCAL_KEY = 'local_key'
    FOREIGN_KEY = 'foreign_key'
    CALCULATED = 'calculated'
    TEMP = 'temp'


class ReadRule(Enum):
    """An Enum class represents JSON Class field's read rule.
    """

    UNLIMITED = 'unlimited'
    NO_READ = 'no_read'


class WriteRule(Enum):
    """An Enum class represents JSON Class field's write rule.
    """

    UNLIMITED = 'unlimited'
    NO_WRITE = 'no_write'
    WRITE_ONCE = 'write_once'
    WRITE_NONNULL = 'write_nonnull'


class Nullability(Enum):
    """An Enum class represents JSON Class field's nullability. This only works
    for collection types and their inner item types.
    """

    UNDEFINED = 'undefined'
    NULLABLE = 'nullable'
    NONNULL = 'nonnull'


class Strictness(Enum):
    """Instance's strictness.
    """

    UNDEFINED = 'undefined'
    STRICT = 'strict'
    UNSTRICT = 'unstrict'


class DeleteRule(Enum):
    """The delete rule of a relationship.
    """

    NULLIFY = 'nullify'
    CASCADE = 'cascade'
    DENY = 'deny'


class EnumInput(Flag):
    NAME = 1
    VALUE = 2
    LOWERCASE_NAME = 4
    NAME_VALUE = NAME | VALUE
    ANYCASE_NAME = NAME | LOWERCASE_NAME
    LOWERCASE_NAME_VALUE = LOWERCASE_NAME | VALUE
    ALL = NAME | VALUE | LOWERCASE_NAME


class EnumOutput(Enum):
    NAME = 'name'
    VALUE = 'value'
    LOWERCASE_NAME = 'lowercase_name'


class Queryability(Enum):
    QUERYABLE = 'queryable'
    UNQUERYABLE = 'unqueryable'


class CopyBehavior(Enum):
    COPY = "copy"
    NOCOPY = "nocopy"


class FDef:
    """The description of a JSONClass field. Some type markers annotate on
    this definition.
    """

    def __init__(self: FDef) -> None:
        self._cdef: Optional[CDef] = None
        self._unresolved: bool = False
        self._unresolved_name: Optional[str] = None
        self._ftype: FType | None = None
        self._fsubtype: FSubtype | None = None
        self._fstore: FStore = FStore.EMBEDDED
        self._getter: Types | Callable | None = None
        self._setter: Types | Callable | None = None
        self._primary: bool = False
        self._index: bool = False
        self._cindex: bool = False
        self._cindex_names: list[str] = []
        self._cunique: bool = False
        self._cunique_names: list[str] = []
        self._unique: bool = False
        self._required: bool = False
        self._raw_enum_class: Optional[type[Enum] | str] = None
        self._enum_class: Optional[type[Enum]] = None
        self._enum_input: Optional[EnumInput] = None
        self._enum_output: Optional[EnumOutput] = None
        self._raw_union_types: Optional[list[Types]] = None
        self._raw_item_types: Optional[Any] = None
        self._raw_inst_types: Optional[str | type[JObject]] = None
        self._resolved_union_types: Optional[list[Types]] = None
        self._resolved_item_types: Optional[Types] = None
        self._inst_cls: Optional[type[JObject]] = None
        self._force_set_on_save: bool = False
        self._foreign_key: Optional[str] = None
        self._use_join_table: Optional[bool] = None
        self._join_table_cls: Optional[Any] = None
        self._join_table_referrer_key: Optional[str] = None
        self._join_table_referee_key: Optional[str] = None
        self._delete_rule: Optional[DeleteRule] = None
        self._read_rule: ReadRule = ReadRule.UNLIMITED
        self._write_rule: WriteRule = WriteRule.UNLIMITED
        self._collection_nullability: Nullability = Nullability.NULLABLE
        self._item_nullability: Nullability = Nullability.UNDEFINED
        self._strictness: Strictness = Strictness.UNDEFINED
        self._has_eager_modifier: bool = False
        self._has_reset_modifier: bool = False
        self._has_preserialize_modifier: bool = False
        self._requires_operator_assign: bool = False
        self._operator_assign_transformer: Optional[Callable] = None
        self._queryability: Optional[Queryability] = None
        self._copy_behavior: Optional[CopyBehavior] = None
        self._auth_identity: bool = False
        self._auth_by: bool = False
        self._auth_by_checker: Optional[Types] = None

    @property
    def cdef(self: FDef) -> CDef:
        """The class definition which owns this field.
        """
        from .cdef import CDef
        return cast(CDef, self._cdef)

    @property
    def ftype(self: FDef) -> FType:
        """The field's type.
        """
        self._resolve_if_needed()
        return cast(FType, self._ftype)

    @property
    def fsubtype(self: FDef) -> FSubtype | None:
        """The field's subtype.
        """
        self._resolve_if_needed()
        return self._fsubtype

    @property
    def fstore(self: FDef) -> FStore:
        """The field's storage.
        """
        self._resolve_if_needed()
        return self._fstore

    @property
    def getter(self: FDef) -> Types | Callable | None:
        """The calculated field's getter.
        """
        self._resolve_if_needed()
        return self._getter

    @property
    def setter(self: FDef) -> Types | Callable | None:
        """The calculated field's setter.
        """
        self._resolve_if_needed()
        return self._setter

    # primary key

    @property
    def primary(self: FDef) -> bool:
        """Whether this field is a primary field. A class can only has one
        primary field.
        """
        self._resolve_if_needed()
        return self._primary

    # database modifiers

    @property
    def index(self: FDef) -> bool:
        """Whether perform single database index on this field. This is marked
        for ORM implementers.
        """
        self._resolve_if_needed()
        return self._index

    @property
    def cindex(self: FDef) -> bool:
        """Whether perform compound database index on this field. This is
        marked for ORM implementers.
        """
        self._resolve_if_needed()
        return self._cindex

    @property
    def cindex_names(self: FDef) -> list[str]:
        """The compound index name for this index field.
        """
        self._resolve_if_needed()
        return self._cindex_names

    @property
    def unique(self: FDef) -> bool:
        """Whether this field's value is unique. This is marked for ORM
        implementers.
        """
        self._resolve_if_needed()
        return self._unique

    @property
    def cunique(self: FDef) -> bool:
        """Whether perform compound database index on this field. This is
        marked for ORM implementers.
        """
        self._resolve_if_needed()
        return self._cunique

    @property
    def cunique_names(self: FDef) -> list[str]:
        """The compound index name for this index field.
        """
        self._resolve_if_needed()
        return self._cunique_names

    @property
    def required(self: FDef) -> bool:
        """Whether this field is required. This is marked for ORM
        implementers.
        """
        self._resolve_if_needed()
        return self._required

    # enum marks

    @property
    def raw_enum_class(self: FDef) -> Optional[type[Enum] | str]:
        """The raw enum class.
        """
        self._resolve_if_needed()
        return self._raw_enum_class

    @property
    def enum_class(self: FDef) -> type[Enum]:
        """The class of the enum.
        """
        self._resolve_if_needed()
        if self._enum_class is not None:
            return self._enum_class
        if isinstance(self._raw_enum_class, str):
            ecls = self.cdef.jconf.cgraph.fetch_enum(self._raw_enum_class)
            self._enum_class = ecls
        else:
            self._enum_class = self._raw_enum_class
        return cast(type[Enum], self._enum_class)

    @property
    def enum_input(self: FDef) -> Optional[EnumInput]:
        """The allowed input types of this enum class.
        """
        self._resolve_if_needed()
        return self._enum_input

    @property
    def enum_output(self: FDef) -> Optional[EnumOutput]:
        """The output type of this enum class.
        """
        self._resolve_if_needed()
        return self._enum_output

    @property
    def raw_union_types(self: FDef) -> Optional[list[Types]]:
        """The raw union types of this union field.
        """
        self._resolve_if_needed()
        return self._raw_union_types

    # subtypes

    @property
    def raw_item_types(self: FDef) -> Optional[Any]:
        """The raw item types of this collection field.
        """
        self._resolve_if_needed()
        return self._raw_item_types

    @property
    def item_types(self: FDef) -> Types:
        """The item types of this collection field.
        """
        from .types import Types
        self._resolve_if_needed()
        if self._raw_item_types is None:
            return cast(Types, None)
        if self._resolved_item_types is not None:
            return self._resolved_item_types
        self._resolved_item_types = rtypes(self.raw_item_types)
        self._resolved_item_types.fdef._cdef = self.cdef
        self._resolved_item_types = rnamedtypes(
            self._resolved_item_types,
            self.cdef.jconf.cgraph,
            self.cdef.name)
        if self._resolved_item_types.fdef.item_nullability == Nullability.UNDEFINED:
            self._resolved_item_types = self._resolved_item_types.required
        return self._resolved_item_types

    @property
    def raw_inst_types(self: FDef) -> Optional[str | type[JObject]]:
        """The raw instance types of this instance field.
        """
        self._resolve_if_needed()
        return self._raw_inst_types

    @property
    def inst_cls(self: FDef) -> Optional[type[JObject]]:
        """The instance class of this field.
        """
        self._resolve_if_needed()
        if self._inst_cls:
            return self._inst_cls
        if self._raw_inst_types is None:
            return None
        if isjsonclass(self._raw_inst_types):
            self._inst_cls = cast(type[JObject], self._raw_inst_types)
        cgraph = self.cdef.jconf.cgraph
        inst_cls = cgraph.fetch(self._raw_inst_types).cls
        self._inst_cls = inst_cls
        return self._inst_cls

    @property
    def force_set_on_save(self: FDef) -> bool:
        """Whether force set on save."""
        self._resolve_if_needed()
        return self._force_set_on_save

    # relationship

    @property
    def foreign_key(self: FDef) -> Optional[str]:
        """The foreign key of the relationship.
        """
        self._resolve_if_needed()
        return self._foreign_key

    @property
    def use_join_table(self: FDef) -> Optional[bool]:
        """Whether this reference uses join table.
        """
        self._resolve_if_needed()
        return self._use_join_table

    @property
    def join_table_cls(self: FDef) -> Optional[Any]:
        """The join table class of the relationship.
        """
        self._resolve_if_needed()
        return self._join_table_cls

    @property
    def join_table_referrer_key(self: FDef) -> Optional[str]:
        """The referrer key of the join table.
        """
        self._resolve_if_needed()
        return self._join_table_referrer_key

    @property
    def join_table_referee_key(self: FDef) -> Optional[str]:
        """The referee key of the join table.
        """
        self._resolve_if_needed()
        return self._join_table_referee_key

    @property
    def delete_rule(self: FDef) -> Optional[DeleteRule]:
        """The delete rule of this relationship.
        """
        self._resolve_if_needed()
        return self._delete_rule

    # read write rule

    @property
    def read_rule(self: FDef) -> ReadRule:
        """The read rule of this field.
        """
        self._resolve_if_needed()
        return self._read_rule

    @property
    def write_rule(self: FDef) -> WriteRule:
        """The write rule of this field.
        """
        self._resolve_if_needed()
        return self._write_rule

    # collection and collection items null rules

    @property
    def collection_nullability(self: FDef) -> Nullability:
        """The collection nullability of this field.
        """
        self._resolve_if_needed()
        return self._collection_nullability

    @property
    def item_nullability(self: FDef) -> Nullability:
        """The item nullability of this field.
        """
        self._resolve_if_needed()
        return self._item_nullability

    @property
    def strictness(self: FDef) -> Strictness:
        """The strictness of this instance field.
        """
        self._resolve_if_needed()
        return self._strictness

    # special modifier marks

    @property
    def has_eager_modifier(self: FDef) -> bool:
        """Whether there is at least an eager modifier in the chain.
        """
        self._resolve_if_needed()
        return self._has_eager_modifier

    @property
    def has_reset_modifier(self: FDef) -> bool:
        """Whether there is at least an reset modifier in the chain.
        """
        self._resolve_if_needed()
        return self._has_reset_modifier

    @property
    def has_preserialize_modifier(self: FDef) -> bool:
        """Whether there is at least a preserialize modifier in the chain.
        """
        self._resolve_if_needed()
        return self._has_preserialize_modifier

    # operator

    @property
    def requires_operator_assign(self: FDef) -> bool:
        """Whether this field requires an operator assigned.
        """
        self._resolve_if_needed()
        return self._requires_operator_assign

    @property
    def operator_assign_transformer(self: FDef) -> Optional[Callable]:
        """The operator assign transformer of this field.
        """
        self._resolve_if_needed()
        return self._operator_assign_transformer

    @property
    def queryability(self: FDef) -> Queryability:
        """The queryability of this field.
        """
        self._resolve_if_needed()
        return self._queryability or Queryability.QUERYABLE

    @property
    def copy_behavior(self: FDef) -> CopyBehavior:
        """The copy behavior of this field.
        """
        self._resolve_if_needed()
        return self._copy_behavior or CopyBehavior.COPY

    @property
    def auth_identity(self: FDef) -> bool:
        """Whether this field is authorization identity.
        """
        self._resolve_if_needed()
        return self._auth_identity

    @property
    def auth_by(self: FDef) -> bool:
        """Whether this field is authorization checker.
        """
        self._resolve_if_needed()
        return self._auth_by

    @property
    def auth_by_checker(self: FDef) -> Types:
        """The auth by checker of this field.
        """
        from .types import Types
        self._resolve_if_needed()
        return cast(Types, self._auth_by_checker)

    # old reference properties

    @property
    def is_ref(self: FDef) -> bool:
        self._resolve_if_needed()
        if self.fstore in \
                [FStore.LOCAL_KEY, FStore.FOREIGN_KEY]:
            return True
        return False

    @property
    def is_inst(self: FDef) -> bool:
        self._resolve_if_needed()
        if self.ftype == FType.INSTANCE:
            return True
        if self.ftype == FType.LIST:
            if self.item_types.fdef.ftype == FType.INSTANCE:
                return True
        return False

    @property
    def has_linked(self: FDef) -> bool:
        self._resolve_if_needed()
        if self.fstore == FStore.LOCAL_KEY:
            return True
        if self.fstore == FStore.FOREIGN_KEY:
            return True
        if self.ftype == FType.LIST:
            return self.item_types.fdef.has_linked
        return False

    def _resolve_if_needed(self: FDef) -> None:
        if self._unresolved:
            # resolve
            self._resolve()
            self._unresolved = False
            self._unresolved_name = None

    def _resolve(self: FDef) -> None:
        self.cdef._resolve_ref_types_if_needed()

    def __str__(self):
        return '<FDef: ' + str(vars(self)) + '>'



from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.ps_user import (PsUserL, PsUserN, PsUserV, PsUserD, PsUserT,
                                   PsUserE, PsUserE2, PsUserE3, PsUserCV,
                                   PsUserLE, PsUserDI, PsUserDE)


class TestPreserializeModifier(TestCase):

    def test_preserialize_modifier_validates_on_save(self):
        with self.assertRaises(ValidationException) as context:
            user = PsUserN(username='123')
            user._set_on_save()
        exception = context.exception
        self.assertEqual(exception.keypath_messages['updatedAt'], "value required")

    def test_preserialize_modifier_does_not_validate_on_normal_validate(self):
        user = PsUserN(username='123')
        user.validate()

    def test_preserialize_modifier_does_not_raise_if_valid_on_save(self):
        user = PsUserV(username='123')
        user._set_on_save()

    def test_preserialize_modifier_setonsave_is_chained_2(self):
        user = PsUserD(username='123')
        user._set_on_save()
        self.assertEqual(user.updated_at, 4)

    def test_preserialize_modifier_setonsave_is_chained_3(self):
        user = PsUserT(username='123')
        user._set_on_save()
        self.assertEqual(user.updated_at, 5)

    def test_preserialize_modifier_validate_between_chains(self):
        with self.assertRaises(ValidationException) as context:
            user = PsUserE(username='123')
            user._set_on_save()
        exception = context.exception
        self.assertEqual(exception.keypath_messages['updatedAt'], "wrong")

    def test_preserialize_modifier_validate_between_chains_2(self):
        with self.assertRaises(ValidationException) as context:
            user = PsUserE2(username='123')
            user._set_on_save()
        exception = context.exception
        self.assertEqual(exception.keypath_messages['updatedAt'], "wrong")

    def test_preserialize_modifier_validate_after_chains(self):
        with self.assertRaises(ValidationException) as context:
            user = PsUserE3(username='123')
            user._set_on_save()
        exception = context.exception
        self.assertEqual(exception.keypath_messages['updatedAt'], "wrong")

    def test_preserialize_modifier_validate_between_chains_do_not_throw_if_valid(self):
        user = PsUserCV(username='123')
        user._set_on_save()

    def test_preserialize_modifier_should_validate_and_setonsave_inside_list(self):
        user = PsUserL(counts=[123, 456])
        user._set_on_save()
        self.assertEqual(user.counts[0], 124)
        self.assertEqual(user.counts[1], 457)

    def test_preserialize_modifier_should_validate_and_throw_inside_list(self):
        with self.assertRaises(ValidationException) as context:
            user = PsUserLE(counts=[123, 456])
            user._set_on_save()
        exception = context.exception
        self.assertEqual(exception.keypath_messages['counts.0'], "value required")

    def test_preserialize_modifier_should_validate_and_setonsave_inside_dict(self):
        user = PsUserDI(counts={'a': 123, 'b': 456})
        user._set_on_save()
        self.assertEqual(user.counts['a'], 124)
        self.assertEqual(user.counts['b'], 457)

    def test_preserialize_modifier_should_validate_and_throw_inside_dict(self):
        with self.assertRaises(ValidationException) as context:
            user = PsUserDE(counts={'a': 123, 'b': 456})
            user._set_on_save()
        exception = context.exception
        self.assertEqual(exception.keypath_messages['counts.a'], "value required")



"""module for fmt modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..types import Types
    from ..ctx import Ctx


class FmtModifier(Modifier):
    """Fmt modifier formats value when presented."""

    def __init__(self, formatter: Callable | Types) -> None:
        super().__init__()
        if callable(formatter):
            params_len = len(signature(formatter).parameters)
            if params_len > 2 or params_len < 1:
                raise ValueError('not a valid formatter callable')
        self.formatter = formatter

    def tojson(self, ctx: Ctx) -> Any:
        from ..types import Types
        if isinstance(self.formatter, Types):
            return self.formatter.modifier.transform(ctx)
        if ctx.val is None:
            return None
        params_len = len(signature(self.formatter).parameters)
        if params_len == 1:
            return self.formatter(ctx.val)
        elif params_len == 2:
            return self.formatter(ctx.val, ctx)



from __future__ import annotations
from jsonclasses import jsonclass, types
from tests.classes.article import Article


@jsonclass
class Author:
    name: str
    articles: list[Article] = types.listof('Article').linkedby('author') \
                                   .required



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperSample:
    digits: Optional[str] = types.str.randomdigits(8)
    alnums: Optional[str] = types.str.randomalnums(8)
    alnumpunc: Optional[str] = types.str.randomalnumpuncs(8)



from unittest import TestCase
from datetime import date
from jsonclasses.excs import ValidationException
from tests.classes.simple_book import SimpleBook
from tests.classes.simple_deadline import SimpleDeadline
from tests.classes.simple_article import SimpleArticle
from tests.classes.article import Article
from tests.classes.author import Author
from tests.classes.default_dict import DefaultDict


class TestSet(TestCase):

    def test_set_without_arguments_wont_change_anything(self):
        book = SimpleBook(name='Thao Bvê', published=False)
        book.set()
        self.assertEqual(book._data_dict,
                         {'name': 'Thao Bvê', 'published': False})

    def test_set_with_keyed_arguments_updates_value(self):
        book = SimpleBook(name='Thao Bvê', published=False)
        book.set(name='Thao Boê')
        self.assertEqual(book._data_dict,
                         {'name': 'Thao Boê', 'published': False})

    def test_set_set_multiple_values_at_once(self):
        book = SimpleBook(name='Thao Boê', published=False)
        book.set(name='Thao Bɛ', published=True)
        self.assertEqual(book._data_dict,
                         {'name': 'Thao Bɛ', 'published': True})

    def test_set_returns_self_and_is_chained(self):
        book = SimpleBook(name='Thao Boê', published=False)
        book.set(name='C').set(name='P').set(name='T').set(name='B')
        self.assertEqual(book._data_dict, {'published': False, 'name': 'B'})

    def test_set_triggers_transform(self):
        deadline = SimpleDeadline()
        deadline.set(ended_at='2020-02-04')
        self.assertEqual(
            deadline._data_dict,
            {
                'ended_at': date.fromisoformat('2020-02-04'),
                'message': None
            }
        )

    def test_set_sets_back_value_to_none(self):
        deadline = SimpleDeadline()
        deadline.set(ended_at='2020-02-04').set(ended_at=None)
        self.assertEqual(
            deadline._data_dict,
            {'ended_at': None, 'message': None})

    def test_set_auto_convert_camelcase_keys_into_snakecase(self):
        deadline = SimpleDeadline()
        deadline.set(**{'endedAt': '2020-02-04'})
        self.assertEqual(
            deadline._data_dict,
            {'ended_at': date(2020, 2, 4), 'message': None})

    def test_set_does_not_accept_undefined_keys_by_default(self):
        article = SimpleArticle(title='Ngê Tu Ngê')
        with self.assertRaises(ValidationException) as context:
            article.set(makcêcê='Tu Ngê Dzu Ngê')
        self.assertTrue(len(context.exception.keypath_messages) == 1)
        self.assertEqual(context.exception.keypath_messages['makcêcê'],
                         "key is not allowed")

    def test_set_accepts_object_list(self):
        article = Article(title='Khi Sit', content='Ua Sim Lai Tsa Tiu E Tsai')
        author = Author(name='Hun')
        author.set(articles=[article])
        self.assertEqual(author.articles, [article])

    def test_set_accepts_object(self):
        author = Author(name='Kieng')
        article = Article(title='E Sai',
                          content='Tsê Tioh Si Kim Sieng Ua Ê Tsuê Ai')
        article.set(author=author)
        self.assertEqual(author, article.author)

    def test_set_accepts_nested_keypaths_for_instances(self):
        author = Author(name='Kieng')
        article = Article(**{'title': 'E Sai',
                             'content': 'Tsê Tioh Si Kim Sieng Ua Ê Tsuê Ai',
                             'author': author})
        article.set(**{'author.name': 'abc.def'})
        self.assertEqual(article.author.name, 'abc.def')

    def test_set_accepts_nested_keypaths_for_instances_in_lists(self):
        article = Article(title='T', content='C')
        author = Author(**{'name': 'Kieng', 'articles': [article]})
        author.set(**{'articles.0.title': 'QQQQQ'})
        self.assertEqual(author.articles[0].title, 'QQQQQ')

    def test_set_accepts_nested_keypaths_for_value_in_dicts(self):
        dct = DefaultDict()
        dct.set(**{'value.b': '3', 'value.c': '4'})
        self.assertEqual(dct.value, {'a': '1', 'b': '3', 'c': '4'})

    def test_set_accepts_nested_keypaths_for_value_in_dicts_with_list_indexing(self):
        dct = DefaultDict()
        dct.set(**{'value[b]': '1'})
        self.assertEqual(dct.value, {'a': '1', 'b': '1'})

    def test_set_accepts_nested_keypaths_for_values_in_dicts_with_list_indexing(self):
        dct = DefaultDict()
        dct.set(**{'value[b]': '1', 'value[c]': '2'})
        self.assertEqual(dct.value, {'a': '1', 'b': '1', 'c': '2'})

    def test_set_accepts_nested_keypaths_for_instances_with_list_indexing(self):
        author = Author(name='Kieng')
        article = Article(**{'title': 'E Sai',
                             'content': 'Tsê Tioh Si Kim Sieng Ua Ê Tsuê Ai',
                             'author': author})
        article.set(**{'author[name]': 'abca.def'})
        self.assertEqual(article.author.name, 'abca.def')

    def test_set_accepts_nested_keypaths_for_instances_with_multiple_list_indexing(self):
        article = Article(title='T', content='C')
        author = Author(**{'name': 'Kieng', 'articles': [article]})
        author.set(**{'articles[0][title]': 'QQQQQ'})
        self.assertEqual(author.articles[0].title, 'QQQQQ')



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class NonnullUser:
    name: str = types.str.writenonnull.required
    nickname: str = types.str.writenonnull.default('KuiPêkBvang').required



"""module for lt modifier."""
from __future__ import annotations
from typing import Callable, Union, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class LtModifier(Modifier):
    """Lt modifier validates value aganinst max value."""

    def __init__(self, lt_value: int | float | Callable | Types):
        self.lt_value = lt_value

    def validate(self, ctx: Ctx) -> None:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        if is_number and ctx.val >= self.resolve_param(self.lt_value, ctx):
            ctx.raise_vexc(f'value is not less than {self.resolve_param(self.lt_value, ctx)}')



from __future__ import annotations
from unittest import TestCase
from datetime import date, datetime
from jsonclasses.excs import ValidationException
from tests.classes.simple_deadline import SimpleDeadline


class TestDate(TestCase):

    def test_date_is_date_after_assigned(self):
        deadline = SimpleDeadline(ended_at=date(2020, 11, 20))
        self.assertEqual(deadline._data_dict,
                         {'ended_at': date(2020, 11, 20),
                          'message': None})

    def test_date_converts_datetime_into_date(self):
        deadline = SimpleDeadline(ended_at=datetime(2020, 6, 30, 1, 1, 1))
        self.assertEqual(deadline.ended_at, date(2020, 6, 30))

    def test_datetime_converts_date_str_into_date(self):
        deadline = SimpleDeadline(ended_at='2020-11-20')
        self.assertEqual(deadline.ended_at, date(2020, 11, 20))
        self.assertEqual(type(deadline.ended_at), date)

    def test_datetime_converts_datetime_str_into_date(self):
        deadline = SimpleDeadline(ended_at='2020-11-20T03:03:03.333Z')
        self.assertEqual(deadline.ended_at, date(2020, 11, 20))

    def test_date_raises_if_value_is_not_date(self):
        deadline = SimpleDeadline(ended_at=True)
        with self.assertRaises(ValidationException) as context:
            deadline.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['endedAt'],
                         "value is not date")

    def test_date_is_datetime_str_when_tojson(self):
        deadline = SimpleDeadline(ended_at='2020-11-20T03:03:03.333Z')
        self.assertEqual(deadline.tojson(),
                         {'endedAt': '2020-11-20T00:00:00.000Z'})



"""module for before modifier."""
from __future__ import annotations
from typing import Callable, Union, TYPE_CHECKING
from datetime import date, datetime
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class BeforeModifier(Modifier):
    """Before modifier validates date against before date."""

    def __init__(self, point: date | datetime | Callable | Types) -> None:
        self.point = point

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        value = ctx.val
        point = self.resolve_param(self.point, ctx)
        if type(value) is date:
            value = datetime.combine(ctx.val, datetime.min.time())
        if type(point) is date:
            point = datetime.combine(self.resolve_param(self.point, ctx), datetime.min.time())
        if point is None:
            return ctx.val
        if value >= point:
            ctx.raise_vexc('value is too late')



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from classes.super_number import SuperNumber

class TestGte(TestCase):

    def test_gte_doesnt_raise_if_float_value_is_greater_than_arg(self):
        n = SuperNumber(ff_gte=5.6)
        n.validate()

    def test_gte_doesnt_raise_if_float_value_is_equal_to_arg(self):
        n = SuperNumber(ff_gte=5.5)
        n.validate()

    def test_gte_raises_if_float_value_is_less_than_arg(self):
        n = SuperNumber(ff_gte=5.4)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ffGte'],
                         "value is not greater than or equal 5.5")

    def test_gte_doesnt_raise_if_int_value_is_greater_than_arg(self):
        n = SuperNumber(if_gte=6)
        n.validate()

    def test_gte_doesnt_raise_if_int_value_is_equal_to_arg(self):
        n = SuperNumber(if_gte=5)
        n.validate()

    def test_gte_raises_if_int_value_is_less_than_arg(self):
        n = SuperNumber(if_gte=4)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ifGte'],
                         "value is not greater than or equal 5")

    def test_gte_doesnt_raise_if_callable_value_is_greater_than_arg(self):
        n = SuperNumber(c_gte=6)
        n.validate()

    def test_gte_doesnt_raise_if_callable_value_is_equal_to_arg(self):
        n = SuperNumber(c_gte=5)
        n.validate()

    def test_gte_raises_if_callable_value_is_less_than_arg(self):
        n = SuperNumber(c_gte=4)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['cGte'],
                         "value is not greater than or equal 5")

    def test_gte_doesnt_raise_if_types_value_is_greater_than_arg(self):
        n = SuperNumber(t_gte=6)
        n.validate()

    def test_gte_doesnt_raise_if_types_value_is_equatto_arg(self):
        n = SuperNumber(t_gte=5)
        n.validate()

    def test_gte_raises_if_types_value_is_less_than_arg(self):
        n = SuperNumber(t_gte=4)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['tGte'],
                         "value is not greater than or equal 5")



"""module for linkto modifier."""
from ..fdef import DeleteRule, FType, FDef, FStore, Nullability
from .modifier import Modifier


class LinkToModifier(Modifier):
    """Link to modifier marks a field which is a local key."""

    def define(self, fdef: FDef) -> None:
        fdef._fstore = FStore.LOCAL_KEY
        if fdef._delete_rule is None:
            fdef._delete_rule = DeleteRule.NULLIFY
        if fdef._ftype == FType.LIST:
            if fdef._collection_nullability is Nullability.UNDEFINED:
                fdef._collection_nullability = Nullability.NONNULL



from __future__ import annotations
from datetime import datetime, date
from unittest import TestCase
from tests.classes.simple_order import SimpleOrder
from tests.classes.default_float import DefaultFloat
from tests.classes.simple_book import SimpleBook
from tests.classes.default_date import DefaultDate
from tests.classes.default_datetime import DefaultDatetime
from tests.classes.default_str import DefaultStr
from tests.classes.default_list import DefaultList
from tests.classes.default_dict import DefaultDict
from tests.classes.linked_default_user import LinkedDefaultUser
from tests.classes.linked_default_post import LinkedDefaultPost


class TestDefault(TestCase):

    def test_default_value_appears_in_int(self):
        simple_order = SimpleOrder()
        self.assertEqual(simple_order.quantity, 1)

    def test_default_value_appears_in_float(self):
        default_float = DefaultFloat()
        self.assertEqual(default_float.value, 1.5)

    def test_default_value_appears_in_str(self):
        default_str = DefaultStr()
        self.assertEqual(default_str.value, 'abc')

    def test_default_value_appears_in_bool(self):
        book = SimpleBook()
        self.assertEqual(book.published, False)

    def test_default_value_appears_in_datetime(self):
        default_datetime = DefaultDatetime()
        self.assertEqual(default_datetime.value, datetime(2000, 1, 20, 0, 0))

    def test_default_value_appears_in_date(self):
        default_date = DefaultDate()
        self.assertEqual(default_date.value, date(2000, 1, 20))

    def test_default_value_appears_in_list(self):
        default_list = DefaultList()
        self.assertEqual(default_list.value, ['1', '2', '3'])

    def test_default_value_appears_in_dict(self):
        default_dict = DefaultDict()
        self.assertEqual(default_dict.value, {'a': '1', 'b': '2'})

    def test_default_value_appears_in_linked_jsonobjects(self):
        user = LinkedDefaultUser(
            posts=[LinkedDefaultPost(), LinkedDefaultPost()])
        self.assertEqual(user.name, 'Default User')
        self.assertEqual(user.posts[0].name, 'Untitled')
        self.assertEqual(user.posts[1].name, 'Untitled')



"""module for sub modifier."""
from __future__ import annotations
from typing import Any, Callable, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class SubModifier(Modifier):
    """Sub modifier subs number value."""

    def __init__(self, by: int | float | Callable | Types):
        self.by = by

    def transform(self, ctx: Ctx) -> Any:
        return ctx.val - self.resolve_param(self.by, ctx) if type(ctx.val) is int or type(ctx.val) is float else ctx.val



"""module for nullify modifier."""
from ..fdef import FDef, DeleteRule
from .modifier import Modifier


class NullifyModifier(Modifier):
    """Nullify modifier marks a relationship's delete rule as nullify."""

    def define(self, fdef: FDef) -> None:
        fdef._delete_rule = DeleteRule.NULLIFY



"""module for temp modifier."""
from .modifier import Modifier
from ..fdef import FStore, FDef


class TempModifier(Modifier):
    """A temp modifier marks a field as temporary field. Value of temporary
    field is never written to database. As long as database write happens,
    temporary fields' values are set to None.
    """

    def define(self, fdef: FDef) -> None:
        fdef._fstore = FStore.TEMP



from __future__ import annotations
from jsonclasses.excs import ValidationException
from tests.classes.super_suffix import SuperSuffix
from unittest import TestCase


class TestIsSuffixOf(TestCase):

    def test_is_suffix_of_validates_str_val_is_suffix_of_str_param(self):
        s_iso = SuperSuffix(s_iso='on')
        s_iso.validate()

    def test_is_suffix_of_validates_list_val_is_suffix_of_list_param(self):
        l_iso = SuperSuffix(l_iso=['ff'])
        l_iso.validate()

    def test_is_suffix_of_raises_if_str_val_is_not_suffix_of_str_param(self):
        s_iso = SuperSuffix(s_iso='py')
        with self.assertRaises(ValidationException) as context:
            s_iso.validate()
        self.assertEqual(context.exception.keypath_messages['sIso'],
                         "suffix is not found")

    def test_is_suffix_of_raises_if_list_val_is_not_suffix_of_list_param(self):
        l_iso = SuperSuffix(l_iso=['ee'])
        with self.assertRaises(ValidationException) as context:
            l_iso.validate()
        self.assertEqual(context.exception.keypath_messages['lIso'],
                         "suffix is not found")

    def test_is_suffix_of_validates_str_val_is_suffix_of_callable_param(self):
        c_iso = SuperSuffix(c_iso='on')
        c_iso.validate()

    def test_is_suffix_of_validates_list_val_is_suffix_of_callable_param(self):
        c_l_iso = SuperSuffix(c_l_iso=['ff'])
        c_l_iso.validate()

    def test_is_suffix_of_raises_if_str_val_is_not_suffix_of_callable_param(self):
        c_iso = SuperSuffix(c_iso='py')
        with self.assertRaises(ValidationException) as context:
            c_iso.validate()
        self.assertEqual(context.exception.keypath_messages['cIso'],
                         "suffix is not found")

    def test_is_suffix_of_raises_if_list_val_is_not_suffix_of_callable_param(self):
        c_l_iso = SuperSuffix(c_l_iso=['ee'])
        with self.assertRaises(ValidationException) as context:
            c_l_iso.validate()
        self.assertEqual(context.exception.keypath_messages['cLIso'],
                         "suffix is not found")

    def test_is_suffix_of_validates_str_val_is_suffix_of_callable_param(self):
        t_iso = SuperSuffix(t_iso='on')
        t_iso.validate()

    def test_is_suffix_of_validates_list_val_is_suffix_of_callable_param(self):
        t_l_iso = SuperSuffix(t_l_iso=['ff'])
        t_l_iso.validate()

    def test_is_suffix_of_raises_if_str_val_is_not_suffix_of_callable_param(self):
        t_iso = SuperSuffix(t_iso='py')
        with self.assertRaises(ValidationException) as context:
            t_iso.validate()
        self.assertEqual(context.exception.keypath_messages['tIso'],
                         "suffix is not found")

    def test_is_suffix_of_raises_if_list_val_is_not_suffix_of_callable_param(self):
        t_l_iso = SuperSuffix(t_l_iso=['ee'])
        with self.assertRaises(ValidationException) as context:
            t_l_iso.validate()
        self.assertEqual(context.exception.keypath_messages['tLIso'],
                         "suffix is not found")



"""module for date modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from datetime import date, datetime
from ..fdef import FType
from .type_modifier import TypeModifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class DateModifier(TypeModifier):
    """Date modifier validate value against date type."""

    def __init__(self):
        super().__init__()
        self.cls = date
        self.ftype = FType.DATE

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        elif isinstance(ctx.val, str):
            try:
                return date.fromisoformat(ctx.val[:10])
            except ValueError:
                ctx.raise_vexc('wrong date format')
        elif type(ctx.val) == datetime:
            return date(ctx.val.year,
                        ctx.val.month,
                        ctx.val.day)
        else:
            return ctx.val

    def tojson(self, ctx: Ctx) -> Any:
        return None if ctx.val is None else self._jsondate(ctx.val)

    def _jsondate(self, d: date) -> str:
        return d.isoformat() + 'T00:00:00.000Z'



"""module for this modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ThisModifier(Modifier):
    """Get the owner object of this field.
    """

    def transform(self, ctx: Ctx) -> Any:
        return ctx.owner



"""module for securepw modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from re import compile, VERBOSE, search
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

securepw_regex = compile(r'''(
    ^(?=.*[A-Z])
    (?=.*[!@#$&*`~()\-_+=\[\]{}:;'",<>.?\\|/])
    (?=.*[0-9])
    (?=.*[a-z])
    .*$
    )''', VERBOSE)

class SecurepwModifier(Modifier):
    """Securepw modifier checks password is secure."""

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, str) and search(securepw_regex, ctx.val) is None:
            ctx.raise_vexc('value is not secure password')



"""module for referrer modifier."""
from ..fdef import FDef
from .modifier import Modifier


class ReferrerModifier(Modifier):
    """Readwrite modifier marks a reference field with a referrer name."""

    def __init__(self, referrer_key: str) -> None:
        self.referrer_key = referrer_key

    def define(self, fdef: FDef) -> None:
        fdef._join_table_referrer_key = self.referrer_key



"""module for match modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Callable
from re import search
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class MatchModifier(Modifier):
    """Match modifier validates value against pattern."""

    def __init__(self, pattern: str | Callable | Types) -> None:
        self.pattern = pattern

    def validate(self, ctx: Ctx) -> None:
        pattern = self.resolve_param(self.pattern, ctx)
        if isinstance(ctx.val, str) and search(pattern, ctx.val) is None:
            ctx.raise_vexc(f'value does not match \'{pattern}\'')



"""module for tobool modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ToBoolModifier(Modifier):
    """ToBool Modifier transforms value into a bool"""

    def transform(self, ctx: Ctx) -> Any:
        return bool(ctx.val)



"""module for nullable modifier."""
from ..fdef import FDef, Nullability
from .modifier import Modifier


class NullableModifier(Modifier):
    """A nullable modifier marks a collection item field to be nullable."""

    def define(self, fdef: FDef) -> None:
        fdef._item_nullability = Nullability.NULLABLE



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_match import SimpleMatch

class TestMatch(TestCase):

    def test_match_does_not_raise_if_matched(self):
        s = SimpleMatch(m="asdsd")
        s.validate()

    def test_match_does_not_raise_if_matched_with_callable_match_str(self):
        s = SimpleMatch(cm="asdsd")
        s.validate()

    def test_match_does_not_raise_if_matched_with_types_match_str(self):
        s = SimpleMatch(tm="asdsd")
        s.validate()

    def test_match_raise_if_it_is_not_matched(self):
        s = SimpleMatch(m="ffsdsd")
        with self.assertRaises(ValidationException) as context:
            s.validate()
        self.assertEqual(context.exception.keypath_messages['m'],
                         "value does not match '^a.*'")



"""module for strict modifier."""
from .modifier import Modifier
from ..fdef import FDef, Strictness


class StrictModifier(Modifier):
    """A strict modifier marks object to disallow undefined keys."""

    def define(self, fdef: FDef) -> None:
        fdef._strictness = Strictness.STRICT



from __future__ import annotations
from unittest import TestCase
from tests.classes.linked_profile import LinkedProfile
from tests.classes.linked_user import LinkedUser
from tests.classes.linked_author import LinkedAuthor
from tests.classes.linked_article import LinkedArticle
from tests.classes.linked_customer import LinkedCustomer
from tests.classes.linked_product import LinkedProduct


class TestAssign(TestCase):

    def test_assign_sets_other_sides_single_link(self):
        user = LinkedUser(name='U')
        profile = LinkedProfile(name='P')
        user.profile = profile
        self.assertEqual(user.profile, profile)
        self.assertEqual(profile.user, user)
        user = LinkedUser(name='U')
        profile = LinkedProfile(name='P')
        profile.user = user
        self.assertEqual(user.profile, profile)
        self.assertEqual(profile.user, user)

    def test_assign_unsets_other_sides_single_link(self):
        user = LinkedUser(name='U')
        profile = LinkedProfile(name='P')
        user.profile = profile
        user.profile = None
        self.assertEqual(user.profile, None)
        self.assertEqual(profile.user, None)
        self.assertEqual(user._unlinked_objects['profile'], [profile])
        self.assertEqual(profile._unlinked_objects['user'], [user])
        user = LinkedUser(name='U')
        profile = LinkedProfile(name='P')
        user.profile = profile
        profile.user = None
        self.assertEqual(user.profile, None)
        self.assertEqual(profile.user, None)
        self.assertEqual(user._unlinked_objects['profile'], [profile])
        self.assertEqual(profile._unlinked_objects['user'], [user])

    def test_assign_resets_other_sides_single_link(self):
        user = LinkedUser(name='U')
        profile = LinkedProfile(name='P')
        user.profile = profile
        user.profile = None
        user.profile = profile
        self.assertEqual(user.profile, profile)
        self.assertEqual(profile.user, user)
        self.assertEqual(user._unlinked_objects['profile'], [])
        self.assertEqual(profile._unlinked_objects['user'], [])

    def test_assign_sets_other_sides_single_to_multiple_link(self):
        article1 = LinkedArticle(name='A1')
        article2 = LinkedArticle(name='A2')
        author = LinkedAuthor(name='Author', articles=[article1])
        article2.author = author
        self.assertEqual(article2.author, author)
        self.assertEqual(author.articles, [article1, article2])
        article1 = LinkedArticle(name='A1')
        article2 = LinkedArticle(name='A2')
        author = LinkedAuthor(name='Author')
        author.articles = [article1, article2]
        self.assertEqual(article1.author, author)
        self.assertEqual(article2.author, author)
        self.assertEqual(author.articles, [article1, article2])

    def test_assign_unsets_other_sides_single_to_multiple_link(self):
        article1 = LinkedArticle(name='A1')
        article2 = LinkedArticle(name='A2')
        author = LinkedAuthor(name='Author', articles=[article1])
        article2.author = author
        article2.author = None
        self.assertEqual(article2.author, None)
        self.assertEqual(author.articles, [article1])
        self.assertEqual(article2._unlinked_objects['author'], [author])
        self.assertEqual(author._unlinked_objects['articles'], [article2])
        article1 = LinkedArticle(name='A1')
        article2 = LinkedArticle(name='A2')
        author = LinkedAuthor(name='Author')
        author.articles = [article1, article2]
        author.articles = []
        self.assertEqual(article1.author, None)
        self.assertEqual(article2.author, None)
        self.assertEqual(author.articles, [])
        self.assertEqual(article1._unlinked_objects['author'], [author])
        self.assertEqual(article2._unlinked_objects['author'], [author])
        self.assertEqual(author._unlinked_objects['articles'],
                         [article1, article2])

    def test_assign_resets_other_sides_single_to_multiple_link(self):
        article1 = LinkedArticle(name='A1')
        article2 = LinkedArticle(name='A2')
        author = LinkedAuthor(name='Author', articles=[article1])
        article2.author = author
        article1.author = None
        article2.author = None
        author.articles = [article1, article2]
        self.assertEqual(article1.author, author)
        self.assertEqual(article2.author, author)
        self.assertEqual(author.articles, [article1, article2])
        self.assertEqual(article1._unlinked_objects['author'], [])
        self.assertEqual(article2._unlinked_objects['author'], [])
        self.assertEqual(author._unlinked_objects['articles'], [])

    def test_assign_sets_other_sides_multiple_to_multiple_link(self):
        c1 = LinkedCustomer(name='C1')
        c2 = LinkedCustomer(name='C2')
        p1 = LinkedProduct(name='P1')
        p2 = LinkedProduct(name='P2')
        p1.customers.extend([c1, c2])
        p2.customers.extend([c1, c2])
        self.assertEqual(c1.products, [p1, p2])
        self.assertEqual(c2.products, [p1, p2])
        self.assertEqual(p1.customers, [c1, c2])
        self.assertEqual(p2.customers, [c1, c2])

    def test_assign_unsets_other_sides_multiple_to_multiple_link(self):
        c1 = LinkedCustomer(name='C1')
        c2 = LinkedCustomer(name='C2')
        p1 = LinkedProduct(name='P1')
        p2 = LinkedProduct(name='P2')
        p1.customers.extend([c1, c2])
        p2.customers.extend([c1, c2])
        p1.customers = []
        p2.customers = []
        self.assertEqual(c1.products, [])
        self.assertEqual(c2.products, [])
        self.assertEqual(p1.customers, [])
        self.assertEqual(p2.customers, [])
        self.assertEqual(c1._unlinked_objects['products'], [p1, p2])
        self.assertEqual(c2._unlinked_objects['products'], [p1, p2])
        self.assertEqual(p1._unlinked_objects['customers'], [c1, c2])
        self.assertEqual(p2._unlinked_objects['customers'], [c1, c2])

    def test_assign_resets_other_sides_multiple_to_multiple_link(self):
        c1 = LinkedCustomer(name='C1')
        c2 = LinkedCustomer(name='C2')
        p1 = LinkedProduct(name='P1')
        p2 = LinkedProduct(name='P2')
        p1.customers.extend([c1, c2])
        p2.customers.extend([c1, c2])
        p1.customers = []
        p2.customers = []
        p1.customers.extend([c1, c2])
        p2.customers.extend([c1, c2])
        self.assertEqual(c1.products, [p1, p2])
        self.assertEqual(c2.products, [p1, p2])
        self.assertEqual(p1.customers, [c1, c2])
        self.assertEqual(p2.customers, [c1, c2])
        self.assertEqual(c1._unlinked_objects['products'], [])
        self.assertEqual(c2._unlinked_objects['products'], [])
        self.assertEqual(p1._unlinked_objects['customers'], [])
        self.assertEqual(p2._unlinked_objects['customers'], [])



from __future__ import annotations
from datetime import date, datetime
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleDatetime:
    fmtdt: Optional[datetime] = types.datetime.fmtd("%Y年%m月%d日 %H:%M:%S")
    fmtd: Optional[date] = types.date.fmtd("%Y年%m月%d日")
    fmtdts: Optional[str] = types.str.fmtd("%Y年%m月%d日 %H:%M:%S")
    fmtds: Optional[str] = types.str.fmtd("%Y年%m月%d日")

    cfmtdt: Optional[datetime] = types.datetime.fmtd(lambda: "%Y年%m月%d日 %H:%M:%S")
    tfmtdt: Optional[datetime] = types.datetime.fmtd(types.default("%Y年%m月%d日 %H:%M:%S"))



"""module for str modifier."""
from ..fdef import FType
from .type_modifier import TypeModifier


class StrModifier(TypeModifier):
    """Str modifier validates value against str type."""

    def __init__(self):
        super().__init__()
        self.cls = str
        self.ftype = FType.STR



from __future__ import annotations
from unittest import TestCase
from tests.classes.linked_permission import LinkedPermission, LinkedOwner
from jsonclasses.excs import UnauthorizedActionException


class TestIsObj(TestCase):

    def test_is_obj_is_valid_if_object_matches(self):
        owner2 = LinkedOwner(id='2')
        permission = LinkedPermission(owner=owner2)
        permission.opby(owner2)
        permission._can_read_check()
        owner1 = LinkedOwner(id='1')
        permission = LinkedPermission(owner=owner2)
        permission.opby(owner1)
        self.assertRaises(UnauthorizedActionException, permission._can_read_check)

    def test_is_obj_is_valid_if_object_id_matches_object(self):
        owner2 = LinkedOwner(id='2')
        permission = LinkedPermission(ownerId='2')
        permission.opby(owner2)
        permission._can_read_check()
        owner1 = LinkedOwner(id='1')
        permission = LinkedPermission(ownerId='2')
        permission.opby(owner1)
        self.assertRaises(UnauthorizedActionException, permission._can_read_check)



from __future__ import annotations
from datetime import datetime
from jsonclasses import jsonclass, types


@jsonclass
class PsUserN:
    username: str
    updated_at: datetime = types.datetime.setonsave(lambda: None).required


@jsonclass
class PsUserV:
    username: str
    updated_at: datetime = types.datetime.setonsave(datetime.now).required


@jsonclass
class PsUserD:
    username: str
    updated_at: int = types.datetime.setonsave(lambda: 2) \
                           .setonsave(lambda x: x * 2)


@jsonclass
class PsUserT:
    username: str
    updated_at: int = types.datetime.setonsave(lambda: 2) \
                           .setonsave(lambda x: x * 2) \
                           .setonsave(lambda x: x + 1)


@jsonclass
class PsUserE:
    username: str
    updated_at: int = types.datetime.setonsave(lambda: 2) \
                           .validate(lambda x: "wrong") \
                           .setonsave(lambda x: x * 2) \
                           .setonsave(lambda x: x + 1)


@jsonclass
class PsUserE2:
    username: str
    updated_at: int = types.datetime.setonsave(lambda: 2) \
                           .setonsave(lambda x: x * 2) \
                           .validate(lambda x: "wrong") \
                           .setonsave(lambda x: x + 1)


@jsonclass
class PsUserE3:
    username: str
    updated_at: int = types.datetime.setonsave(lambda: 2) \
                           .setonsave(lambda x: x * 2) \
                           .setonsave(lambda x: x + 1) \
                           .validate(lambda x: "wrong")


@jsonclass
class PsUserCV:
    username: str
    updated_at: int = types.datetime \
                           .setonsave(lambda: 2).validate(lambda x: None) \
                           .setonsave(lambda x: x * 2).validate(lambda x: None) \
                           .setonsave(lambda x: x + 1).validate(lambda x: None)


@jsonclass
class PsUserL:
    counts: list[int] = types.listof(types.int.setonsave(lambda s: s + 1))


@jsonclass
class PsUserLE:
    counts: list[int] = types.listof(types.int.setonsave(lambda s: None))


@jsonclass
class PsUserDI:
    counts: dict[str, int] = types.dictof(types.int.setonsave(lambda s: s + 1))


@jsonclass
class PsUserDE:
    counts: dict[str, int] = types.dictof(types.int.setonsave(lambda s: None))



"""module for floor modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from math import floor
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class FloorModifier(Modifier):
    """Floor modifier Floor number value."""

    def transform(self, ctx: Ctx) -> Any:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        return floor(ctx.val) if is_number else ctx.val



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass(class_graph='grouptemplate')
class Group:
    id: int = types.int.primary
    name: str
    template: Template = types.objof('Template').linkedby('group').present


@jsonclass(class_graph='grouptemplate')
class Template:
    id: int = types.int.primary
    name: str
    group: Group = types.linkto.objof('Group').required



"""module for index modifier."""
from .modifier import Modifier
from ..fdef import FDef


class IndexModifier(Modifier):
    """Index modifier implies this column should be indexed in database."""

    def __init__(self, unique: bool, index_name: str | None = None) -> None:
        self.unique = unique
        self.index_name = index_name

    def define(self, fdef: FDef) -> None:
        if self.index_name:
            if self.unique:
                fdef._cunique = True
                fdef._cunique_names.append(self.index_name)
            else:
                fdef._cindex = True
                fdef._cindex_names.append(self.index_name)
        else:
            if self.unique:
                fdef._unique = True
            else:
                fdef._index = True



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_append import SuperAppend

class TestAppend(TestCase):

    def test_append_adds_str_into_the_end_of_original_str(self):
        i_str = SuperAppend(s='99999999')
        self.assertEqual(i_str.s, '999999993432')

    def test_append_adds_int_into_the_end_of_original_list(self):
        l_list = SuperAppend(l=['a', 'c', 's', 'fsa'])
        self.assertEqual(l_list.l, ['a', 'c', 's', 'fsa', '7788'])

    def test_append_appends_callable_str_into_original_str(self):
        i_str = SuperAppend(cs='99999999')
        self.assertEqual(i_str.cs, '999999993432')

    def test_append_appends_callable_any_into_original_list(self):
        l_list = SuperAppend(cl=['a', 'c', 's', 'fsa'])
        self.assertEqual(l_list.cl, ['a', 'c', 's', 'fsa', '7788'])

    def test_append_appends_types_str_into_original_str(self):
        i_str = SuperAppend(ts='99999999')
        self.assertEqual(i_str.ts, '999999993432')

    def test_append_appends_types_any_into_original_list(self):
        l_list = SuperAppend(tl=['a', 'c', 's', 'fsa'])
        self.assertEqual(l_list.tl, ['a', 'c', 's', 'fsa', '7788'])



from __future__ import annotations
from importlib import import_module
from subprocess import check_call
from sys import executable


def check_package(module_name: str, info: tuple[str, str]) -> None:
    try:
        import_module(module_name)
    except:
        raise ModuleNotFoundError(f'module {module_name} is not found. please install {info[0]}{info[1]}')


def install_package(info: tuple[str, str]) -> None:
    param = f'{info[0]}{info[1]}'
    check_call([executable, "-m", "pip", "install", param])
    with open('requirements.txt', 'r') as file:
        content = file.read()
        lines = content.split('\n')
        has = False
        for line in lines:
            if info[0] in line:
                has = True
        if not has:
            with open('requirements.txt', 'a') as afile:
                afile.write(param + '\n')


def check_and_install_packages(packages: dict[str, (str, str)] | None) -> None:
    for module_name, info in packages.items():
        try:
            check_package(module_name, info)
        except:
            install_package(info)



from __future__ import annotations
from typing import Optional
from datetime import datetime
from jsonclasses import jsonclass


@jsonclass
class SimpleBalance:
    date: Optional[datetime]
    balance: Optional[float]



from __future__ import annotations
from tests.classes.simple_eq import SimpleNeq
from jsonclasses.excs import ValidationException
from unittest import TestCase


class TestNeq(TestCase):

    def test_neq_does_not_raise_if_is_not_equal(self):
        n = SimpleNeq(neq_value="dsa")
        n.validate()

    def test_neq_raises_if_is_equal(self):
        n = SimpleNeq(neq_value="dsadsa")
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(context.exception.keypath_messages['neqValue'],
                         "value is not unequal")

    def test_neq_does_not_raise_if_unequal_with_callable_value(self):
        e = SimpleNeq(cneq_value="dsa")
        e.validate()

    def test_neq_raises_if_is_equal_with_callable_value(self):
        e = SimpleNeq(cneq_value="dsadsa")
        with self.assertRaises(ValidationException) as context:
            e.validate()
        self.assertEqual(context.exception.keypath_messages['cneqValue'],
                         "value is not unequal")

    def test_neq_does_not_raise_if_unequal_with_types_value(self):
        e = SimpleNeq(tneq_value="dsa")
        e.validate()

    def test_neq_raises_if_is_equal_with_types_value(self):
        e = SimpleNeq(tneq_value="dsadsa")
        with self.assertRaises(ValidationException) as context:
            e.validate()
        self.assertEqual(context.exception.keypath_messages['tneqValue'],
                         "value is not unequal")



"""module for modifier modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from inspect import signature
from ..fdef import FDef
from ..pkgutils import check_and_install_packages
if TYPE_CHECKING:
    from ..ctx import Ctx


class Modifier:
    """Abstract and base class for modifiers."""

    def packages(self) -> dict[str, (str, str)] | None:
        return None

    def check_packages(self) -> None:
        check_and_install_packages(self.packages())

    def resolve_param(self, param: Any, ctx: Ctx) -> Any:
        from ..types import Types
        if isinstance(param, Types):
            newctx = ctx.nval(None)
            return param.modifier.transform(newctx)
        elif callable(param):
            params_len = len(signature(param).parameters)
            if params_len == 0:
                return param()
            elif params_len == 1:
                return param(ctx.owner)
            elif params_len == 2:
                return param(ctx.owner, ctx)
            else:
                raise ValueError('not a valid parameter')
        else:
            return param

    def define(self, fdef: FDef) -> None:
        """A hook and chance for modifier to update field description."""

    def validate(self, ctx: Ctx) -> None:
        """Validate the validity of the object."""

    def transform(self, ctx: Ctx) -> Any:
        """Transform raw input object into JSON Class acceptable object."""
        return ctx.val

    def tojson(self, ctx: Ctx) -> Any:
        """Transform JSON Class object and fields into JSON dict and values."""
        return ctx.val

    def serialize(self, ctx: Ctx) -> Any:
        """A chance for modifiers to update the object's value before the
        value is serialized into the database. This is only triggered for
        objects which are modified and have fields to write to the database.

        Unmodified objects won't cause serialize to trigger. JSON Classes which
        are not subclasses of ORMObject won't trigger this.
        """
        return ctx.val



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_type import SuperType


class TestToFloat(TestCase):

    def test_tofloat_transforms_str_into_float(self):
        f = SuperType(tf='-113.0')
        self.assertEqual(f.tf, -113.0)

    def test_tofloat_transforms_bool_into_float(self):
        f = SuperType(tf=False)
        self.assertEqual(f.tf, 0.0)



"""module for default modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class DefaultModifier(Modifier):
    """Default modifier assigns field a default value if value is `None`."""

    def __init__(self, default: Any | Callable | Types) -> None:
        self.default = default

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is not None:
            return ctx.val
        if callable(self.default):
            return self.default()
        else:
            return self.resolve_param(self.default, ctx)



"""This module defines the `JField` named tuple. This data structure
records the detailed information of a JSON class field.
"""
from __future__ import annotations
from jsonclasses.jobject import JObject
from jsonclasses.fdef import FStore, FType
from typing import Any, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from .types import Types
    from .fdef import FDef
    from .cdef import CDef
    from .modifiers import ChainedModifier


class JField:
    """The detailed field information of a JSON class field. This includes
    field names in different circumstances, assigned default value, field
    description and main modifier.
    """

    def __init__(
            self: JField, cdef: CDef, name: str, default: Any, types: Types
        ) -> None:
        self._cdef = cdef
        self._name = name
        self._default = default
        self._types = types
        self._resolved_foreign = False
        self._foreign_cdef = None
        self._foreign_field = None
        self._foreign_fname = None
        self._resolved_foreign_class = False
        self._foreign_class = None

    @property
    def cdef(self: JField) -> CDef:
        """The class definition on which this field is defined.
        """
        return self._cdef

    @property
    def name(self: JField) -> str:
        """The name of the field in Python.
        """
        return self._name

    @property
    def json_name(self: JField) -> str:
        """The name of the field when converted into JSON dict.
        """
        return self.cdef.jconf.output_key_strategy(self._name)

    @property
    def default(self: JField) -> Any:
        """The default value that user assigned with equal sign.
        """
        return self._default

    @property
    def types(self: JField) -> Types:
        """The user defined field types definition or auto generated types
        definition.
        """
        return self._types

    @property
    def fdef(self: JField) -> FDef:
        """The detailed field definition defined with the types chain.
        """
        return self._types.fdef

    @property
    def modifier(self: JField) -> ChainedModifier:
        """The chained modifier of the field.
        """
        return self._types.modifier

    @property
    def foreign_cdef(self: JField) -> Optional[CDef]:
        self.cdef._resolve_ref_types_if_needed()
        if self._resolved_foreign:
            return self._foreign_cdef
        self._resolve_foreign()
        return self._foreign_cdef

    @property
    def foreign_fname(self: JField) -> Optional[str]:
        self.cdef._resolve_ref_types_if_needed()
        if self._resolved_foreign:
            return self._foreign_fname
        self._resolve_foreign()
        return self._foreign_fname

    @property
    def foreign_field(self: JField) -> Optional[JField]:
        self.cdef._resolve_ref_types_if_needed()
        if self._resolved_foreign:
            return self._foreign_field
        self._resolve_foreign()
        return self._foreign_field

    @property
    def foreign_class(self: JField) -> Optional[type[JObject]]:
        self.cdef._resolve_ref_types_if_needed()
        if self._resolved_foreign_class:
            return self._foreign_class
        if self.fdef.ftype == FType.INSTANCE:
            self._foreign_class = self.fdef.inst_cls
        elif self.fdef.ftype == FType.LIST:
            self._foreign_class = self.fdef.item_types.fdef.inst_cls
        self._resolved_foreign_class = True
        return self._foreign_class

    @property
    def is_primary(self) -> bool:
        return self.fdef.primary

    @property
    def is_inst_field(self) -> bool:
        return self.fdef.ftype == FType.INSTANCE

    @property
    def is_list_field(self) -> bool:
        return self.fdef.ftype == FType.LIST

    @property
    def is_list_inst_field(self) -> bool:
        if not self.is_list_field:
            return False
        t = self.fdef.item_types
        if t.fdef.raw_inst_types is not None:
            return True
        return False

    @property
    def is_foreign_key_store(self) -> bool:
        return self.fdef.fstore == FStore.FOREIGN_KEY

    @property
    def is_local_key_store(self) -> bool:
        return self.fdef.fstore == FStore.LOCAL_KEY

    @property
    def is_foreign_one_ref(self) -> bool:
        return self.is_inst_field and self.is_foreign_key_store

    @property
    def is_foreign_many_ref(self) -> bool:
        return self.is_list_field and self.is_foreign_key_store

    @property
    def is_local_one_ref(self) -> bool:
        return self.is_inst_field and self.is_local_key_store

    @property
    def is_local_many_ref(self) -> bool:
        return self.is_list_field and self.is_local_key_store

    @property
    def is_join_table_ref(self) -> bool:
        return self.fdef.use_join_table is True

    @property
    def ref_name(self) -> str:
        return self.cdef.jconf.ref_name_strategy(self)

    def _resolve_foreign(self: JField) -> None:
        self._do_resolve_foreign()
        self._resolved_foreign = True

    def _do_resolve_foreign(self: JField) -> None:
        from .fdef import FStore
        if self.fdef.fstore not in [FStore.LOCAL_KEY, FStore.FOREIGN_KEY]:
            return None
        scls = self.fdef.cdef.cls
        slocal = self.fdef.fstore == FStore.LOCAL_KEY
        fcls = self.foreign_class
        if self.fdef.ftype == FType.INSTANCE:
            stype = 'inst'
        elif self.fdef.ftype == FType.LIST:
            stype = 'list'
        if not fcls:
            return
        self._foreign_cdef = fcls.cdef
        if slocal:
            ffield = fcls.cdef.rfield(scls, None, self.name)
            if ffield:
                self._foreign_field = ffield
                self._foreign_fname = ffield.name
        else:
            fkey = self.fdef.foreign_key
            ffield = fcls.cdef.rfield(scls, fkey, None)
            if ffield:
                self._foreign_field = ffield
                self._foreign_fname = ffield.name



"""module for unresolved modifier."""
from ..fdef import FDef
from .modifier import Modifier


class UnresolvedModifier(Modifier):
    """This modifier contains unresolved names."""

    def __init__(self, arg: str) -> None:
        self.arg = arg

    def define(self, fdef: FDef) -> None:
        fdef._unresolved = True
        fdef._unresolved_name = self.arg
